import math
from typing import Any, Literal

import numpy as np
from numpy.typing import ArrayLike

from paicorelib.framelib.frames import NDArray

from ..coordinate import CoordZXYOffset, coordzxy_to_sign_magnitude
from ..core_defs import LCN_EX
from ..core_defs_v2 import CSCAccelerateMode, DataWidth
from ..core_model_v2 import OfflineCoreRegV2
from ..neuron_model_v2 import (
    OfflineNeuDestInfoV2,
    OfflineNeuFoldedAttrsV2Part1,
    OfflineNeuFoldedAttrsV2Part2,
    OfflineNeuFullAttrsV2Part1,
    OfflineNeuFullAttrsV2Part2,
)
from ..routing_hexa import AERPacketZXYCopy
from .base import FramePackageHeaderV2, FrameV2, get_frame_destV2
from .frame_defs import FrameHeader as FH
from .frame_defs import FramePackageType
from .frame_defs import OfflineConfigFrame1FormatV2 as Off_Cfg1_V2
from .frame_defs import OfflineConfigFrame2FormatV2 as Off_Cfg2_V2
from .frame_defs import OfflineConfigFrame3FormatV2 as Off_Cfg3_V2
from .frame_defs import OfflineControlFrame1FormatV2 as Off_Ctrl1_V2
from .frame_defs import OfflineControlFrame3FormatV2 as Off_Ctrl3_V2
from .frame_defs import OfflineWorkFrame1FormatV2 as Off_Work1_V2
from .types import (
    FRAME_DTYPE,
    PAYLOAD_DATA_DTYPE,
    FrameArrayType,
    LUTActivationType,
    LUTPotentialType,
)
from .utils import _mask, bin_split

__all__ = ["FrameGenV2", "OfflineFrameGenV2"]


class FrameGenV2:
    @staticmethod
    def make_package(
        header: FH,
        pkt_offset: CoordZXYOffset,
        pkt_ncopy: AERPacketZXYCopy,
        start_addr: int,
        packages: FrameArrayType,
    ) -> FrameArrayType:
        pkg_header = FramePackageHeaderV2.make_pkg_header(
            header,
            pkt_offset,
            pkt_ncopy,
            start_addr,
            FramePackageType.CONF_TESTOUT,
            len(packages),
        )
        return np.concatenate([pkg_header.value, packages])


class OfflineFrameGenV2(FrameGenV2):
    LCN_TO_TS_AXON_WIDTHS = (
        (8, 9),
        (7, 10),
        (6, 11),
        (5, 12),
        (4, 13),
        (3, 14),
        (2, 15),
        (1, 16),  # LCN_128X
    )

    @staticmethod
    def gen_config_frame1(
        pkt_offset: CoordZXYOffset,
        core_reg_: OfflineCoreRegV2 | dict[str, Any],
        pkt_ncopy: AERPacketZXYCopy = AERPacketZXYCopy(),
        start_addr: int = 0,
    ) -> FrameArrayType:
        """Generate a configuration frame type I. The number of packages is calculated automatically."""
        F = Off_Cfg1_V2
        core_reg = OfflineCoreRegV2.model_validate(core_reg_, strict=True).model_dump()

        # Convert the test coordzxy in sign-magnitude format
        z, x, y = coordzxy_to_sign_magnitude(
            (core_reg["test_core_xy"], core_reg["test_core_x"], core_reg["test_core_y"])
        )
        test_core_y_h2, test_core_y_l4 = bin_split(y, 4, 2)
        w1 = (
            ((core_reg["snn_ann"] & F.Word1.SNN_ANN_MASK) << F.Word1.SNN_ANN_OFFSET)
            | (
                (core_reg["max_pooling"] & F.Word1.MAX_POOLING_MASK)
                << F.Word1.MAX_POOLING_OFFSET
            )
            | (
                (core_reg["add_potential"] & F.Word1.ADD_POTENTIAL_MASK)
                << F.Word1.ADD_POTENTIAL_OFFSET
            )
            | (
                (core_reg["zero_output"] & F.Word1.ZERO_OUTPUT_MASK)
                << F.Word1.ZERO_OUTPUT_OFFSET
            )
            | (
                (core_reg["input_sign"] & F.Word1.INPUT_SIGN_MASK)
                << F.Word1.INPUT_SIGN_OFFSET
            )
            | (
                (core_reg["input_width"] & F.Word1.INPUT_WIDTH_MASK)
                << F.Word1.INPUT_WIDTH_OFFSET
            )
            | (
                (core_reg["output_sign"] & F.Word1.OUTPUT_SIGN_MASK)
                << F.Word1.OUTPUT_SIGN_OFFSET
            )
            | (
                (core_reg["output_width"] & F.Word1.OUTPUT_WIDTH_MASK)
                << F.Word1.OUTPUT_WIDTH_OFFSET
            )
            | (
                (core_reg["weight_sign"] & F.Word1.WEIGHT_SIGN_MASK)
                << F.Word1.WEIGHT_SIGN_OFFSET
            )
            | (
                (core_reg["weight_width"] & F.Word1.WEIGHT_WIDTH_MASK)
                << F.Word1.WEIGHT_WIDTH_OFFSET
            )
            | ((core_reg["lcn"] & F.Word1.LCN_MASK) << F.Word1.LCN_OFFSET)
            | (
                (core_reg["target_lcn"] & F.Word1.TARGET_LCN_MASK)
                << F.Word1.TARGET_LCN_OFFSET
            )
            | (
                (core_reg["axon_skew"] & F.Word1.AXON_SKEW_MASK)
                << F.Word1.AXON_SKEW_OFFSET
            )
            | (
                (core_reg["neuron_number"] & F.Word1.NEURON_NUMBER_MASK)
                << F.Word1.NEURON_NUMBER_OFFSET
            )
            | ((z & F.Word1.TEST_CORE_XY_MASK) << F.Word1.TEST_CORE_XY_OFFSET)
            | ((x & F.Word1.TEST_CORE_X_MASK) << F.Word1.TEST_CORE_X_OFFSET)
            | (
                (test_core_y_h2 & F.Word1.TEST_CORE_Y_HIGH2_MASK)
                << F.Word1.TEST_CORE_Y_HIGH2_OFFSET
            )
        )
        w2 = (
            (
                (test_core_y_l4 & F.Word2.TEST_CORE_Y_LOW4_MASK)
                << F.Word2.TEST_CORE_Y_LOW4_OFFSET
            )
            | (
                (core_reg["global_send"] & F.Word2.GLOBAL_SEND_MASK)
                << F.Word2.GLOBAL_SEND_OFFSET
            )
            | (
                (core_reg["csc_accelerate"] & F.Word2.CSC_ACCELERATE_MASK)
                << F.Word2.CSC_ACCELERATE_OFFSET
            )
            | (
                (core_reg["global_receive"] & F.Word2.GLOBAL_RECEIVE_MASK)
                << F.Word2.GLOBAL_RECEIVE_OFFSET
            )
            | (
                (core_reg["thread_number"] & F.Word2.THREAD_NUMBER_MASK)
                << F.Word2.THREAD_NUMBER_OFFSET
            )
            | (
                (core_reg["busy_cycle"] & F.Word2.BUSY_CYCLE_MASK)
                << F.Word2.BUSY_CYCLE_OFFSET
            )
            | (
                (core_reg["delay_cycle"] & F.Word2.DELAY_CYCLE_MASK)
                << F.Word2.DELAY_CYCLE_OFFSET
            )
            | (
                (core_reg["width_cycle"] & F.Word2.WIDTH_CYCLE_MASK)
                << F.Word2.WIDTH_CYCLE_OFFSET
            )
        )
        w3 = (
            (
                (core_reg["tick_start"] & F.Word3.TICK_START_MASK)
                << F.Word3.TICK_START_OFFSET
            )
            | (
                (core_reg["tick_duration"] & F.Word3.TICK_DURATION_MASK)
                << F.Word3.TICK_DURATION_OFFSET
            )
            | (
                (core_reg["tick_initial"] & F.Word3.TICK_INITIAL_MASK)
                << F.Word3.TICK_INITIAL_OFFSET
            )
        )

        pkg = np.array([w1, w2, w3], dtype=FRAME_DTYPE)
        return OfflineFrameGenV2.make_package(
            FH.CONFIG_TYPE1, pkt_offset, pkt_ncopy, start_addr, pkg
        )

    @staticmethod
    def gen_config_frame2(
        pkt_offset: CoordZXYOffset,
        potentials: LUTPotentialType,
        activations: LUTActivationType,
        pkt_ncopy: AERPacketZXYCopy = AERPacketZXYCopy(),
        start_addr: int = 0,
    ) -> FrameArrayType:
        """Generate a configuration frame type II. The number of packages is calculated automatically."""
        F = Off_Cfg2_V2
        N_FRAME_PER_LUT_RAM = 256

        assert potentials.size == activations.size
        packages = np.zeros((N_FRAME_PER_LUT_RAM,), dtype=FRAME_DTYPE)

        arr_pot_u64 = potentials.view(np.uint32).astype(FRAME_DTYPE)
        arr_act_u8 = activations.astype(np.uint8)

        packages = ((arr_pot_u64 << F.POTENTIAL_OFFSET) + arr_act_u8).astype(
            FRAME_DTYPE
        )
        return OfflineFrameGenV2.make_package(
            FH.CONFIG_TYPE2, pkt_offset, pkt_ncopy, start_addr, packages
        )

    @staticmethod
    def gen_config_frame3_pkg_header(
        pkt_offset: CoordZXYOffset,
        start_addr: int,
        n_package: int,
        pkt_ncopy: AERPacketZXYCopy = AERPacketZXYCopy(),
    ) -> FrameArrayType:
        """Generate the package header of configuration frame type III."""
        header = FramePackageHeaderV2.make_pkg_header(
            FH.CONFIG_TYPE3,
            pkt_offset,
            pkt_ncopy,
            start_addr,
            FramePackageType.CONF_TESTOUT,
            n_package,
        )
        return header.value

    @staticmethod
    def gen_config_frame3_pkg_neu(
        dest_info: OfflineNeuDestInfoV2 | dict[str, Any],
        full_attrs1: OfflineNeuFullAttrsV2Part1 | dict[str, Any] | None,
        full_attrs2: OfflineNeuFullAttrsV2Part2 | dict[str, Any] | None,
        folded_attrs1: OfflineNeuFoldedAttrsV2Part1 | dict[str, Any] | None,
        folded_attrs2_: (list[OfflineNeuFoldedAttrsV2Part2] | list[dict[str, Any]]),
    ) -> tuple[FrameArrayType, FrameArrayType, FrameArrayType]:
        """Generate three packages of half, full & folded neuron attributes."""
        dest_info = OfflineNeuDestInfoV2.model_validate(
            dest_info, strict=True
        ).model_dump()

        pkg_half_neu = np.array([], dtype=FRAME_DTYPE)
        pkg_full_neu = np.array([], dtype=FRAME_DTYPE)
        pkg_folded_neu = np.array([], dtype=FRAME_DTYPE)

        if full_attrs1 is not None:  # half
            full_attrs1 = OfflineNeuFullAttrsV2Part1.model_validate(
                full_attrs1, strict=True
            ).model_dump()
            pkg_half_neu = OfflineFrameGenV2._gen_pkg_half_neu(dest_info, full_attrs1)

            if full_attrs2 is not None:  # full
                full_attrs2 = OfflineNeuFullAttrsV2Part2.model_validate(
                    full_attrs2, strict=True
                ).model_dump()
                pkg_full_neu = OfflineFrameGenV2._gen_pkg_full_neu(
                    dest_info, full_attrs1, full_attrs2
                )

        if folded_attrs1 is not None and len(folded_attrs2_) > 0:  # folded
            folded_attrs1 = OfflineNeuFoldedAttrsV2Part1.model_validate(
                folded_attrs1, strict=True
            ).model_dump()
            folded_attrs2 = [
                OfflineNeuFoldedAttrsV2Part2.model_validate(
                    attrs2, strict=True
                ).model_dump()
                for attrs2 in folded_attrs2_
            ]
            pkg_folded_neu = OfflineFrameGenV2._gen_pkg_folded_neu(
                folded_attrs1, *folded_attrs2
            )
        elif folded_attrs1 is None and len(folded_attrs2_) == 0:
            pass
        else:
            raise ValueError("attributes of folded neuron are incomplete")

        return pkg_half_neu, pkg_full_neu, pkg_folded_neu

    @staticmethod
    def _gen_pkg_half_neu(
        dest_info: dict[str, Any], half_attrs: dict[str, Any]
    ) -> FrameArrayType:
        F = Off_Cfg3_V2.Full
        weight_skew_h11, weight_skew_l5 = bin_split(half_attrs["weight_skew"], 5, 11)
        # RAM[0][63:0]
        w1 = (
            (
                (weight_skew_l5 & F.Word1.WEIGHT_SKEW_LOW5_MASK)
                << F.Word1.WEIGHT_SKEW_LOW5_OFFSET
            )
            | (
                (half_attrs["weight_address_start"] & F.Word1.WEIGHT_ADDRESS_START_MASK)
                << F.Word1.WEIGHT_ADDRESS_START_OFFSET
            )
            | (
                (half_attrs["weight_address_end"] & F.Word1.WEIGHT_ADDRESS_END_MASK)
                << F.Word1.WEIGHT_ADDRESS_END_OFFSET
            )
            | (
                (half_attrs["output_type"] & F.Word1.OUTPUT_TYPE_MASK)
                << F.Word1.OUTPUT_TYPE_OFFSET
            )
            | (
                (half_attrs["fold_type"] & F.Word1.FOLD_TYPE_MASK)
                << F.Word1.FOLD_TYPE_OFFSET
            )
            | (
                (half_attrs["neuron_type"] & F.Word1.NEURON_TYPE_MASK)
                << F.Word1.NEURON_TYPE_OFFSET
            )
            | ((half_attrs["vjt"] & F.Word1.VJT_MASK) << F.Word1.VJT_OFFSET)
        )
        # RAM[0][127:64]
        w2 = (
            (
                (dest_info["tick_relative"] & F.Word2.TICK_RELATIVE_MASK)
                << F.Word2.TICK_RELATIVE_OFFSET
            )
            | (
                (dest_info["addr_axon"] & F.Word2.ADDR_AXON_MASK)
                << F.Word2.ADDR_AXON_OFFSET
            )
            | (
                (dest_info["addr_core_xy"] & F.Word2.ADDR_CORE_XY_MASK)
                << F.Word2.ADDR_CORE_XY_OFFSET
            )
            | (
                (dest_info["addr_core_x"] & F.Word2.ADDR_CORE_X_MASK)
                << F.Word2.ADDR_CORE_X_OFFSET
            )
            | (
                (dest_info["addr_core_y"] & F.Word2.ADDR_CORE_Y_MASK)
                << F.Word2.ADDR_CORE_Y_OFFSET
            )
            | (
                (dest_info["addr_copy_xy"] & F.Word2.ADDR_COPY_XY_MASK)
                << F.Word2.ADDR_COPY_XY_OFFSET
            )
            | (
                (dest_info["addr_copy_x"] & F.Word2.ADDR_COPY_X_MASK)
                << F.Word2.ADDR_COPY_X_OFFSET
            )
            | (
                (dest_info["addr_copy_y"] & F.Word2.ADDR_COPY_Y_MASK)
                << F.Word2.ADDR_COPY_Y_OFFSET
            )
            | (
                (weight_skew_h11 & F.Word2.WEIGHT_SKEW_HIGH11_MASK)
                << F.Word2.WEIGHT_SKEW_HIGH11_OFFSET
            )
        )
        return np.array([w1, w2], dtype=FRAME_DTYPE)

    @staticmethod
    def _gen_pkg_full_neu(
        dest_info: dict[str, Any],
        full_attrs1: dict[str, Any],
        full_attrs2: dict[str, Any],
    ) -> FrameArrayType:
        F = Off_Cfg3_V2.Full
        pkg_half_neu = OfflineFrameGenV2._gen_pkg_half_neu(dest_info, full_attrs1)
        thres_pos_h12, thres_pos_l20 = bin_split(full_attrs2["threshold_neg"], 20, 12)
        # RAM[1][63:0]
        w3 = (
            (
                (thres_pos_l20 & F.Word3.THRESHOLD_POS_LOW20_MASK)
                << F.Word3.THRESHOLD_POS_LOW20_OFFSET
            )
            | (
                (full_attrs2["lateral_inhibition"] & F.Word3.LATERAL_INHIBITION_MASK)
                << F.Word3.LATERAL_INHIBITION_OFFSET
            )
            | (
                (full_attrs2["leak_multi_sequence"] & F.Word3.LEAK_MULTI_SEQUENCE_MASK)
                << F.Word3.LEAK_MULTI_SEQUENCE_OFFSET
            )
            | (
                (full_attrs2["leak_multi_input"] & F.Word3.LEAK_MULTI_INPUT_MASK)
                << F.Word3.LEAK_MULTI_INPUT_OFFSET
            )
            | (
                (full_attrs2["leak_multi_mode"] & F.Word3.LEAK_MULTI_MODE_MASK)
                << F.Word3.LEAK_MULTI_MODE_OFFSET
            )
            | (
                (full_attrs2["leak_add_mode"] & F.Word3.LEAK_ADD_MODE_MASK)
                << F.Word3.LEAK_ADD_MODE_OFFSET
            )
            | (
                (full_attrs2["leak_tau"] & F.Word3.LEAK_TAU_MASK)
                << F.Word3.LEAK_TAU_OFFSET
            )
            | ((full_attrs2["leak_v"] & F.Word3.LEAK_V_MASK) << F.Word3.LEAK_V_OFFSET)
            | (
                (full_attrs2["weight_compress"] & F.Word3.WEIGHT_COMPRESS_MASK)
                << F.Word3.WEIGHT_COMPRESS_OFFSET
            )
            | (
                (full_attrs2["vjt_initial"] & F.Word3.VJT_INITIAL_MASK)
                << F.Word3.VJT_INITIAL_OFFSET
            )
        )
        # RAM[1][127:64]
        w4 = (
            (
                (full_attrs2["reset_mode"] & F.Word4.RESET_MODE_MASK)
                << F.Word4.RESET_MODE_OFFSET
            )
            | (
                (full_attrs2["reset_v"] & F.Word4.RESET_V_MASK)
                << F.Word4.RESET_V_OFFSET
            )
            | (
                (full_attrs2["threshold_neg_mode"] & F.Word4.THRESHOLD_NEG_MODE_MASK)
                << F.Word4.THRESHOLD_NEG_MODE_OFFSET
            )
            | (
                (full_attrs2["threshold_pos_mode"] & F.Word4.THRESHOLD_POS_MODE_MASK)
                << F.Word4.THRESHOLD_POS_MODE_OFFSET
            )
            | (
                (full_attrs2["threshold_neg"] & F.Word4.THRESHOLD_NEG_MASK)
                << F.Word4.THRESHOLD_NEG_OFFSET
            )
            | (
                (thres_pos_h12 & F.Word4.THRESHOLD_POS_HIGH12_MASK)
                << F.Word4.THRESHOLD_POS_HIGH12_OFFSET
            )
        )
        return np.r_[pkg_half_neu, w3, w4].astype(FRAME_DTYPE)

    @staticmethod
    def _gen_pkg_folded_neu(
        folded_attrs1: dict[str, Any], *folded_attrs2: dict[str, Any]
    ) -> FrameArrayType:
        F = Off_Cfg3_V2.Fold
        fold_skew_y_h9, fold_skew_y_l2 = bin_split(folded_attrs1["fold_skew_y"], 2, 9)
        # RAM[0][63:0]
        w1 = (
            (
                (fold_skew_y_l2 & F.Word1.FOLD_SKEW_Y_LOW2_MASK)
                << F.Word1.FOLD_SKEW_Y_LOW2_OFFSET
            )
            | (
                (folded_attrs1["fold_axon_xy"] & F.Word1.FOLD_AXON_XY_MASK)
                << F.Word1.FOLD_AXON_XY_OFFSET
            )
            | (
                (folded_attrs1["fold_axon_x"] & F.Word1.FOLD_AXON_X_MASK)
                << F.Word1.FOLD_AXON_X_OFFSET
            )
            | (
                (folded_attrs1["fold_axon_y"] & F.Word1.FOLD_AXON_Y_MASK)
                << F.Word1.FOLD_AXON_Y_OFFSET
            )
            | (
                (folded_attrs1["fold_number"] & F.Word1.FOLD_NUMBER_MASK)
                << F.Word1.FOLD_NUMBER_OFFSET
            )
        )
        # RAM[1][127:64]
        w2 = (
            (
                (folded_attrs1["fold_range_xy"] & F.Word2.FOLD_RANGE_XY_MASK)
                << F.Word2.FOLD_RANGE_XY_OFFSET
            )
            | (
                (folded_attrs1["fold_range_x"] & F.Word2.FOLD_RANGE_X_MASK)
                << F.Word2.FOLD_RANGE_X_OFFSET
            )
            | (
                (folded_attrs1["fold_range_y"] & F.Word2.FOLD_RANGE_Y_MASK)
                << F.Word2.FOLD_RANGE_Y_OFFSET
            )
            | (
                (folded_attrs1["fold_skew_xy"] & F.Word2.FOLD_SKEW_XY_MASK)
                << F.Word2.FOLD_SKEW_XY_OFFSET
            )
            | (
                (folded_attrs1["fold_skew_x"] & F.Word2.FOLD_SKEW_X_MASK)
                << F.Word2.FOLD_SKEW_X_OFFSET
            )
            | (
                (fold_skew_y_h9 & F.Word2.FOLD_SKEW_Y_HIGH9_MASK)
                << F.Word2.FOLD_SKEW_Y_HIGH9_OFFSET
            )
        )

        assert len(folded_attrs2) > 0
        v0 = np.array([item["fold_vjt_0"] for item in folded_attrs2], dtype=FRAME_DTYPE)
        v1 = np.array([item["fold_vjt_1"] for item in folded_attrs2], dtype=FRAME_DTYPE)
        v2 = np.array([item["fold_vjt_2"] for item in folded_attrs2], dtype=FRAME_DTYPE)
        v3 = np.array([item["fold_vjt_3"] for item in folded_attrs2], dtype=FRAME_DTYPE)

        # RAM[0][63:0]
        w3 = ((v1 & F.Word3.FOLD_VJT_1_MASK) << F.Word3.FOLD_VJT_1_OFFSET) | (
            (v0 & F.Word3.FOLD_VJT_0_MASK) << F.Word3.FOLD_VJT_0_OFFSET
        )
        # RAM[1][127:64]
        w4 = ((v3 & F.Word4.FOLD_VJT_3_MASK) << F.Word4.FOLD_VJT_3_OFFSET) | (
            (v2 & F.Word4.FOLD_VJT_2_MASK) << F.Word4.FOLD_VJT_2_OFFSET
        )
        v = np.zeros(len(w3) * 2, dtype=FRAME_DTYPE)
        v[0::2] = w3
        v[1::2] = w4

        return np.r_[w1, w2, v].astype(FRAME_DTYPE)

    @staticmethod
    def gen_config_frame3_weight_pkg(
        weight: np.ndarray,
        weight_width: DataWidth | Literal[1, 2, 4, 8],
        input_width: DataWidth | Literal[1, 2, 4, 8],
        csc_compress: bool | CSCAccelerateMode = False,
    ) -> FrameArrayType:
        """Generate weight package for config frame type III."""
        assert weight.ndim == 1

        is_compress = csc_compress != CSCAccelerateMode.DISABLE
        if isinstance(weight_width, DataWidth):
            weight_width = 1 << weight_width.value
        else:
            weight_width = weight_width

        if isinstance(input_width, DataWidth):
            input_width = 1 << input_width.value
        else:
            input_width = input_width

        if is_compress:
            return weight_csc_pack(weight, weight_width, input_width)
        else:
            return weight_dense_pack(weight, weight_width)

    @staticmethod
    def gen_config_frame4(
        pkt_offset: CoordZXYOffset,
        input_array: FrameArrayType,
        pkt_ncopy: AERPacketZXYCopy = AERPacketZXYCopy(),
        start_addr: int = 0,
    ):
        # TODO
        pass

    @staticmethod
    def gen_work_frame1(
        pkt_offset: CoordZXYOffset,
        pkt_ncopy: AERPacketZXYCopy,
        timesteps: ArrayLike,
        axons: ArrayLike,
        target_lcn: LCN_EX,
        data: ArrayLike,
    ) -> FrameArrayType:
        F = Off_Work1_V2
        _data = np.asarray(data, dtype=PAYLOAD_DATA_DTYPE)
        ts = np.asarray(timesteps, dtype=FRAME_DTYPE).ravel()
        ax = np.asarray(axons, dtype=FRAME_DTYPE).ravel()

        if ax.size != ts.size:
            raise ValueError(
                f"the size of axons & timeslots are not equal, {ax.size} != {ts.size}."
            )
        if _data.size != ts.size:
            raise ValueError(
                f"the size of data & timeslots are not equal, {_data.size} != {ts.size}."
            )

        TS_WIDTH, AX_WIDTH = OfflineFrameGenV2.LCN_TO_TS_AXON_WIDTHS[target_lcn.value]
        ts_msb = (ts >> (TS_WIDTH - 1)) & F.TIMESTEP_HIGH7_MASK
        ts_low = ts & _mask(TS_WIDTH - 1)

        ts_ax_addr = (
            (ts_msb << F.TIMESTEP_HIGH7_OFFSET)
            | (ts_low << (F.AXON_ADDR_OFFSET + AX_WIDTH))
            | ((ax & _mask(AX_WIDTH)) << F.AXON_ADDR_OFFSET)
        )

        mask = np.flatnonzero(_data)
        data_u8 = _data.astype(PAYLOAD_DATA_DTYPE)
        frame_dest = get_frame_destV2(FH.WORK_TYPE1, pkt_offset, pkt_ncopy)
        return (frame_dest + ts_ax_addr[mask] + data_u8[mask]).astype(FRAME_DTYPE)

    @staticmethod
    def gen_control_frame1(
        pkt_offset: CoordZXYOffset, pkt_ncopy: AERPacketZXYCopy, n_timestep: int
    ) -> FrameArrayType:
        F = Off_Ctrl1_V2
        if n_timestep > F.NUM_TIMESTEP_MASK:
            raise ValueError(f"'overflow' out of range {F.NUM_TIMESTEP_MASK}")

        return FrameV2(FH.CTRL_TYPE1, pkt_offset, pkt_ncopy, n_timestep).value

    @staticmethod
    def gen_control_frame2(
        pkt_offset: CoordZXYOffset,
        pkt_ncopy: AERPacketZXYCopy,
    ) -> FrameArrayType:
        return FrameV2(FH.CTRL_TYPE2, pkt_offset, pkt_ncopy, 0).value

    @staticmethod
    def gen_complete_frame(
        pkt_offset: CoordZXYOffset,
        pkt_ncopy: AERPacketZXYCopy,
        thread_id: int = 0,
    ) -> FrameArrayType:
        F = Off_Ctrl3_V2
        if thread_id > F.THREAD_ID_MASK:
            raise ValueError(f"'thread_id' out of range {F.THREAD_ID_MASK})")

        return FrameV2(FH.CTRL_TYPE3, pkt_offset, pkt_ncopy, thread_id).value


# class OfflineFrameDecoderV2:
#     @staticmethod
#     def decode_voltage(frame: FrameArrayType, target_lcn: LCN_EX) -> FrameArrayType:
#         pass


def weight_dense_pack(
    weight: np.ndarray, weight_width: Literal[1, 2, 4, 8]
) -> FrameArrayType:
    """Array uncompressed weights in RAM.

    weight[127:0] -> RAM[0][ 63: 0]
                     RAM[0][127:64]
    128/64/32/16 1/2/4/8-bit weights are stored in one address of RAM, little endian.
    """
    weight = weight.astype(np.uint8).ravel()
    align_size = 128 // weight_width
    aligned_size = math.ceil(weight.size / align_size) * align_size
    if (pad := aligned_size - weight.size) > 0:
        weight = np.pad(weight, (0, pad), constant_values=0)

    if weight_width == 8:
        return weight.view(FRAME_DTYPE)

    weights_per_byte = 8 // weight_width
    mask = _mask(weight_width)
    weight = weight.reshape(-1, weights_per_byte)  # (N, 8/ww)

    shifts = weight_width * np.arange(weights_per_byte, dtype=np.uint8)
    # Reduce to u8 (N,) and then view 8*u8 as u64
    reduced = np.bitwise_or.reduce((weight & mask) << shifts, axis=1, dtype=np.uint8)
    return reduced.view(FRAME_DTYPE)


def weight_csc_pack(
    weight: np.ndarray, weight_width: Literal[1, 2, 4, 8], input_width: Literal[1, 2, 4, 8]
) -> FrameArrayType:
    """Arrange compressed weights according to CSC format.

    weight[127:0]: col_indices  +   weight
    1-bit           [127:16]        [ 6:0] stored 7 non-zero 1-bit weights
    2-bit           [127:16]        [13:0] stored 7 non-zero 2-bit weights
    4-bit           [127:32]        [23:0] stored 6 non-zero 4-bit weights
    8-bit           [127:32]        [39:0] stored 5 non-zero 8-bit weights

    NOTE: Non-zero values must be aligned in groups of 7/7/6/5. If all the values of a sparse weight    \
        are non-zero but need alignment, an exception will raise. If there is any 0, the first index    \
        pointing to 0 will be used as padding.
    """
    weight = weight.astype(np.uint8).ravel()
    # #N of non-zero weight stored in a single address of RAM
    N_NONZERO_WEIGHT_PER_ADDR = {1: 7, 2: 7, 4: 6, 8: 5}
    INDICES_ADDR_OFFSET = {1: 16, 2: 16, 4: 32, 8: 48}

    row_indices = np.flatnonzero(weight)
    w_nonzero = weight[row_indices]

    n_nonzero_w_per_addr = N_NONZERO_WEIGHT_PER_ADDR[weight_width]
    n_chunk = math.ceil(row_indices.size / n_nonzero_w_per_addr)

    if (pad := n_chunk * n_nonzero_w_per_addr - row_indices.size) > 0:
        if w_nonzero.size == weight.size:
            # NOTE: If the weight is all non-zero but needed to be padded, it's impossible to align.
            raise ValueError(
                f"the sparse weight cannot be aligned in groups of {n_nonzero_w_per_addr} "
                + "because there are all non-zero values."
            )

        # Get the first index of 0, pad with zeros & set the index pointing to it.
        idx_first_zero = np.where(weight == 0)[0][0]
        w_nonzero = np.pad(w_nonzero, (0, pad), constant_values=0)
        row_indices = np.pad(row_indices, (0, pad), constant_values=idx_first_zero)

    # (chunk, N)
    w_chunks = w_nonzero.reshape(n_chunk, n_nonzero_w_per_addr)
    # Index is no more than u16


    # in csc pack, the indice stored in RAM is the bit offset of the non-zero weight, 
    # which is the original index multiplied by input_width.
    row_indices = row_indices * input_width
    idx_chunks = row_indices.reshape(n_chunk, n_nonzero_w_per_addr).astype(np.uint16)

    w_shifts = weight_width * np.arange(n_nonzero_w_per_addr, dtype=np.uint8)
    idx_shifts = INDICES_ADDR_OFFSET[weight_width] + 16 * np.arange(
        n_nonzero_w_per_addr, dtype=np.uint8
    )
    w_mask = FRAME_DTYPE(_mask(weight_width))
    idx_mask = FRAME_DTYPE(_mask(16))

    # Reduce the weight & indices to (chunk,) u64
    w_reduced_chunk = np.bitwise_or.reduce(
        (w_chunks & w_mask) << w_shifts, axis=1, dtype=FRAME_DTYPE
    )

    n_idx_at_high = 4  # 4 indices will placed at high [127:64]
    idx_chunks_h = idx_chunks[:, -n_idx_at_high:]
    idx_chunks_l = idx_chunks[:, :-n_idx_at_high]
    idx_shifs_h = idx_shifts[-n_idx_at_high:] - 64
    idx_shifs_l = idx_shifts[:-n_idx_at_high]

    idx_reduced_chunk_h = np.bitwise_or.reduce(
        (idx_chunks_h & idx_mask) << idx_shifs_h, axis=1, dtype=FRAME_DTYPE
    )
    idx_reduced_chunk_l = np.bitwise_or.reduce(
        (idx_chunks_l & idx_mask) << idx_shifs_l, axis=1, dtype=FRAME_DTYPE
    )

    result = np.zeros((2 * n_chunk,), dtype=FRAME_DTYPE)
    # Little endian
    result[0::2] = idx_reduced_chunk_l + w_reduced_chunk
    result[1::2] = idx_reduced_chunk_h
    return result


def weight_dense_unpack(
    frames: FrameArrayType,
    weight_width: Literal[1, 2, 4, 8],
    signed: bool,
    original_size: int,
) -> NDArray[np.int8 | np.uint8]:
    w_per_u64 = 64 // weight_width
    mask = _mask(weight_width)

    shifts = weight_width * np.arange(w_per_u64, dtype=np.uint8)
    extracted = ((frames[:, np.newaxis] >> shifts) & mask).ravel()

    # Omit the part of padding
    result = extracted[:original_size]

    if signed and weight_width < 8:
        signbit = 1 << (weight_width - 1)
        result = result.astype(np.int8)
        result[result >= signbit] -= 1 << weight_width

    return result.astype(np.int8 if signed else np.uint8)


def weight_csc_unpack(
    frames: FrameArrayType,
    weight_width: Literal[1, 2, 4, 8],
    signed: bool,
    original_size: int,
) -> NDArray[np.int8 | np.uint8]:
    # #N of non-zero weight stored in a single address of RAM
    N_NONZERO_WEIGHT_PER_ADDR = {1: 7, 2: 7, 4: 6, 8: 5}
    INDICES_ADDR_OFFSET = {1: 16, 2: 16, 4: 32, 8: 48}
    n_nonzero_w_per_addr = N_NONZERO_WEIGHT_PER_ADDR[weight_width]

    assert frames.size % 2 == 0

    w_shifts = weight_width * np.arange(n_nonzero_w_per_addr, dtype=np.uint8)
    idx_shifts = INDICES_ADDR_OFFSET[weight_width] + 16 * np.arange(
        n_nonzero_w_per_addr, dtype=np.uint8
    )
    w_mask = _mask(weight_width)
    idx_mask = _mask(16)

    # (chunk,)
    chunk_low64 = frames[0::2]
    chunk_high64 = frames[1::2]

    weights = ((chunk_low64[:, np.newaxis] >> w_shifts[np.newaxis, :]) & w_mask).astype(
        np.uint8
    )

    n_idx_at_high = 4  # 4 indices will placed at high [127:64]
    idx_shifs_h = idx_shifts[-n_idx_at_high:] - 64
    idx_shifs_l = idx_shifts[:-n_idx_at_high]
    indices_h = (chunk_high64[:, np.newaxis] >> idx_shifs_h) & idx_mask
    indices_l = (chunk_low64[:, np.newaxis] >> idx_shifs_l) & idx_mask
    # Little endian
    indices = np.hstack([indices_l, indices_h])

    if signed and weight_width < 8:
        signbit = 1 << (weight_width - 1)
        weights = weights.astype(np.int8)
        weights[weights >= signbit] -= 1 << weight_width

    result = np.zeros(original_size, dtype=np.int8 if signed else np.uint8)
    result[indices] = weights
    return result



from collections.abc import Sequence
from typing import Any, ClassVar, overload

import numpy as np
from numpy.typing import ArrayLike

from ..coordinate import ChipCoord, Coord
from ..coordinate import ReplicationId as RId
from ..core_defs import LCN_EX, CoreType
from ..core_defs import WeightWidth as WW
from ..core_model import OfflineCoreReg as OffCoreReg
from ..core_model import OnlineCoreReg as OnCoreReg
from ..neuron_model import OfflineNeuAttrs as OffNeuAttrs
from ..neuron_model import OfflineNeuDestInfo as OffNeuDestInfo
from ..neuron_model import OnlineNeuAttrs as OnNeuAttrs
from ..neuron_model import OnlineNeuDestInfo as OnNeuDestInfo
from ..routing_defs import _rid_unset
from .frames import (
    OfflineConfigFrame1,
    OfflineConfigFrame2,
    OfflineConfigFrame3,
    OfflineConfigFrame4,
    OfflineTestInFrame1,
    OfflineTestInFrame2,
    OfflineTestInFrame3,
    OfflineTestInFrame4,
    OfflineTestOutFrame1,
    OfflineTestOutFrame2,
    OfflineTestOutFrame3,
    OfflineTestOutFrame4,
    OfflineWorkFrame1,
    OfflineWorkFrame2,
    OfflineWorkFrame3,
    OfflineWorkFrame4,
    OnlineConfigFrame1,
    OnlineConfigFrame2,
    OnlineConfigFrame3,
    OnlineConfigFrame4,
    OnlineTestInFrame1,
    OnlineTestInFrame2,
    OnlineTestInFrame3,
    OnlineTestInFrame4,
    OnlineTestOutFrame1,
    OnlineTestOutFrame2,
    OnlineTestOutFrame3,
    OnlineTestOutFrame4,
    OnlineWorkFrame1_1,
    OnlineWorkFrame1_2,
    OnlineWorkFrame1_3,
    OnlineWorkFrame1_4,
    OnlineWorkFrame2,
    OnlineWorkFrame3,
    OnlineWorkFrame4,
)
from .types import (
    FRAME_DTYPE,
    PAYLOAD_DATA_DTYPE,
    FrameArrayType,
    IntScalarType,
    LUTDataType,
    PayloadDataType,
)

__all__ = ["OfflineFrameGen", "OnlineFrameGen", "ChipFrameGen"]


class OfflineFrameGen:
    """Offline frame generator."""

    @staticmethod
    def gen_config_frame1(
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        random_seed: IntScalarType,
    ) -> OfflineConfigFrame1:
        return OfflineConfigFrame1(chip_coord, core_coord, rid, int(random_seed))

    @staticmethod
    def gen_config_frame2(
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        core_reg: OffCoreReg | dict[str, Any],
    ) -> OfflineConfigFrame2:
        return OfflineConfigFrame2(chip_coord, core_coord, rid, core_reg)

    @overload
    @staticmethod
    def gen_config_frame3(
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        ram_start_addr: int,
        n_neuron: int,
        attrs: OffNeuAttrs,
        dest_info: OffNeuDestInfo,
        repeat: int = 1,
    ) -> OfflineConfigFrame3: ...

    @overload
    @staticmethod
    def gen_config_frame3(
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        ram_start_addr: int,
        n_neuron: int,
        attrs: dict[str, Any],
        dest_info: dict[str, Any],
        repeat: int = 1,
    ) -> OfflineConfigFrame3: ...

    @staticmethod
    def gen_config_frame3(
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        ram_start_addr: int,
        n_neuron: int,
        attrs: OffNeuAttrs | dict[str, Any],
        dest_info: OffNeuDestInfo | dict[str, Any],
        repeat: int = 1,
    ) -> OfflineConfigFrame3:
        return OfflineConfigFrame3(
            chip_coord,
            core_coord,
            rid,
            ram_start_addr,
            n_neuron,
            attrs,
            dest_info,
            repeat,
        )

    @staticmethod
    def gen_config_frame4(
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        ram_start_addr: int,
        n_data_package: int,
        weight_ram: FrameArrayType,
    ) -> OfflineConfigFrame4:
        return OfflineConfigFrame4(
            chip_coord, core_coord, rid, ram_start_addr, n_data_package, weight_ram
        )

    @staticmethod
    def gen_testin_frame1(
        chip_coord: ChipCoord, core_coord: Coord, rid: RId, /
    ) -> OfflineTestInFrame1:
        return OfflineTestInFrame1(chip_coord, core_coord, rid)

    @staticmethod
    def gen_testout_frame1(
        test_chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        random_seed: IntScalarType,
    ) -> OfflineTestOutFrame1:
        return OfflineTestOutFrame1(test_chip_coord, core_coord, rid, int(random_seed))

    @staticmethod
    def gen_testin_frame2(
        chip_coord: ChipCoord, core_coord: Coord, rid: RId, /
    ) -> OfflineTestInFrame2:
        return OfflineTestInFrame2(chip_coord, core_coord, rid)

    @staticmethod
    def gen_testout_frame2(
        test_chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        core_reg: OffCoreReg | dict[str, Any],
    ) -> OfflineTestOutFrame2:
        return OfflineTestOutFrame2(test_chip_coord, core_coord, rid, core_reg)

    @staticmethod
    def gen_testin_frame3(
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        ram_start_addr: int,
        n_package: int,
    ) -> OfflineTestInFrame3:
        return OfflineTestInFrame3(
            chip_coord, core_coord, rid, ram_start_addr, n_package
        )

    @overload
    @staticmethod
    def gen_testout_frame3(
        test_chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        ram_start_addr: int,
        n_neuron: int,
        attrs: OffNeuAttrs | dict[str, Any],
        dest_info: OffNeuDestInfo | dict[str, Any],
        lcn_ex: LCN_EX,
        weight_width: WW,
    ) -> OfflineTestOutFrame3: ...

    @overload
    @staticmethod
    def gen_testout_frame3(
        test_chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        ram_start_addr: int,
        n_neuron: int,
        attrs: OffNeuAttrs | dict[str, Any],
        dest_info: OffNeuDestInfo | dict[str, Any],
        *,
        repeat: int,
    ) -> OfflineTestOutFrame3: ...

    @staticmethod
    def gen_testout_frame3(
        test_chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        ram_start_addr: int,
        n_neuron: int,
        attrs: OffNeuAttrs | dict[str, Any],
        dest_info: OffNeuDestInfo | dict[str, Any],
        lcn_ex: LCN_EX | None = None,
        weight_width: WW | None = None,
        *,
        repeat: int | None = None,
    ) -> OfflineTestOutFrame3:
        if lcn_ex is not None and weight_width is not None:
            repeat = 1 << (lcn_ex + weight_width)
        else:
            assert repeat is not None

        return OfflineTestOutFrame3(
            test_chip_coord,
            core_coord,
            rid,
            ram_start_addr,
            n_neuron,
            attrs,
            dest_info,
            repeat,
        )

    @staticmethod
    def gen_testin_frame4(
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        ram_start_addr: int,
        n_package: int,
    ) -> OfflineTestInFrame4:
        return OfflineTestInFrame4(
            chip_coord, core_coord, rid, ram_start_addr, n_package
        )

    @staticmethod
    def gen_testout_frame4(
        test_chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        ram_start_addr: int,
        n_data_package: int,
        weight_ram: FrameArrayType,
    ) -> OfflineTestOutFrame4:
        return OfflineTestOutFrame4(
            test_chip_coord, core_coord, rid, ram_start_addr, n_data_package, weight_ram
        )

    @staticmethod
    def gen_work_frame1(
        one_input_node: dict[str, Any], data: ArrayLike
    ) -> FrameArrayType:
        """Generate the common part of the input spike frames by given the info of one input node.

        Args:
            one_input_node: a dictionary of a single input node that points to offline cores.
            data: the input data.
        """
        common_frame_dest = OfflineWorkFrame1._frame_dest_reorganized(one_input_node)
        _data = np.asarray(data, dtype=PAYLOAD_DATA_DTYPE)

        return OfflineFrameGen.gen_work_frame1_fast(common_frame_dest, _data)

    @staticmethod
    def gen_work_frame1_fast(
        frame_dest_info: FrameArrayType, data: PayloadDataType
    ) -> FrameArrayType:
        if frame_dest_info.size != data.size:
            raise ValueError(
                f"the size of frame dest info & data are not equal, {frame_dest_info.size} != {data.size}."
            )

        mask = np.flatnonzero(data)
        return frame_dest_info[mask] + data[mask]

    @staticmethod
    def gen_work_frame2(chip_coord: ChipCoord, /, n_sync: int) -> OfflineWorkFrame2:
        return OfflineWorkFrame2(chip_coord, n_sync)

    @staticmethod
    def gen_work_frame3(chip_coord: ChipCoord) -> OfflineWorkFrame3:
        return OfflineWorkFrame3(chip_coord)

    @staticmethod
    def gen_work_frame4(chip_coord: ChipCoord) -> OfflineWorkFrame4:
        return OfflineWorkFrame4(chip_coord)


class OnlineFrameGen:
    """Online frame generator."""

    @staticmethod
    def gen_config_frame1(
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        lut: LUTDataType | None = None,
    ) -> OnlineConfigFrame1:
        return OnlineConfigFrame1(chip_coord, core_coord, rid, lut)

    @staticmethod
    def gen_config_frame2(
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        core_reg: OnCoreReg | dict[str, Any],
    ) -> OnlineConfigFrame2:
        return OnlineConfigFrame2(chip_coord, core_coord, rid, core_reg)

    @overload
    @staticmethod
    def gen_config_frame3(
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        ram_start_addr: int,
        n_neuron: int,
        attrs: OnNeuAttrs,
        dest_info: OnNeuDestInfo,
        weight_width: WW,
    ) -> OnlineConfigFrame3: ...

    @overload
    @staticmethod
    def gen_config_frame3(
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        ram_start_addr: int,
        n_neuron: int,
        attrs: dict[str, Any],
        dest_info: dict[str, Any],
        weight_width: WW,
    ) -> OnlineConfigFrame3: ...

    @staticmethod
    def gen_config_frame3(
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        ram_start_addr: int,
        n_neuron: int,
        attrs: OnNeuAttrs | dict[str, Any],
        dest_info: OnNeuDestInfo | dict[str, Any],
        weight_width: WW,
    ) -> OnlineConfigFrame3:
        return OnlineConfigFrame3(
            chip_coord,
            core_coord,
            rid,
            ram_start_addr,
            n_neuron,
            attrs,
            dest_info,
            weight_width,
        )

    @staticmethod
    def gen_config_frame4(
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        ram_start_addr: int,
        n_data_package: int,
        weight_ram: FrameArrayType,
    ) -> OnlineConfigFrame4:
        return OnlineConfigFrame4(
            chip_coord, core_coord, rid, ram_start_addr, n_data_package, weight_ram
        )

    @staticmethod
    def gen_testin_frame1(
        chip_coord: ChipCoord, core_coord: Coord, rid: RId
    ) -> OnlineTestInFrame1:
        return OnlineTestInFrame1(chip_coord, core_coord, rid)

    @staticmethod
    def gen_testout_frame1(
        test_chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        lut: LUTDataType | None = None,
    ) -> OnlineTestOutFrame1:
        return OnlineTestOutFrame1(test_chip_coord, core_coord, rid, lut)

    @staticmethod
    def gen_testin_frame2(
        chip_coord: ChipCoord, core_coord: Coord, rid: RId
    ) -> OnlineTestInFrame2:
        return OnlineTestInFrame2(chip_coord, core_coord, rid)

    @staticmethod
    def gen_testout_frame2(
        test_chip_coord: Coord,
        core_coord: Coord,
        rid: RId,
        /,
        core_reg: OnCoreReg | dict[str, Any],
    ) -> OnlineTestOutFrame2:
        return OnlineTestOutFrame2(test_chip_coord, core_coord, rid, core_reg)

    @staticmethod
    def gen_testin_frame3(
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        ram_start_addr: int,
        n_package: int,
    ) -> OnlineTestInFrame3:
        return OnlineTestInFrame3(
            chip_coord, core_coord, rid, ram_start_addr, n_package
        )

    @overload
    @staticmethod
    def gen_testout_frame3(
        test_chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        ram_start_addr: int,
        n_neuron: int,
        attrs: OnNeuAttrs,
        dest_info: OnNeuDestInfo,
        weight_width: WW,
    ) -> OnlineTestOutFrame3: ...

    @overload
    @staticmethod
    def gen_testout_frame3(
        test_chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        ram_start_addr: int,
        n_neuron: int,
        attrs: dict[str, Any],
        dest_info: dict[str, Any],
        weight_width: WW,
    ) -> OnlineTestOutFrame3: ...

    @staticmethod
    def gen_testout_frame3(
        test_chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        ram_start_addr: int,
        n_neuron: int,
        attrs: OnNeuAttrs | dict[str, Any],
        dest_info: OnNeuDestInfo | dict[str, Any],
        weight_width: WW,
    ) -> OnlineTestOutFrame3:
        return OnlineTestOutFrame3(
            test_chip_coord,
            core_coord,
            rid,
            ram_start_addr,
            n_neuron,
            attrs,
            dest_info,
            weight_width,
        )

    @staticmethod
    def gen_testin_frame4(
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        ram_start_addr: int,
        n_package: int,
    ) -> OnlineTestInFrame4:
        return OnlineTestInFrame4(
            chip_coord, core_coord, rid, ram_start_addr, n_package
        )

    @staticmethod
    def gen_testout_frame4(
        test_chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        /,
        ram_start_addr: int,
        n_data_package: int,
        weight_ram: FrameArrayType,
    ) -> OnlineTestOutFrame4:
        return OnlineTestOutFrame4(
            test_chip_coord, core_coord, rid, ram_start_addr, n_data_package, weight_ram
        )

    @staticmethod
    def gen_work_frame1_1(
        one_input_node: dict[str, Any], data: ArrayLike
    ) -> FrameArrayType:
        """Generate the common part of the input spike frames by given the info of one input node.

        Args:
            one_input_node: a dictionary of a single input node that points to online cores.
        """
        common_frame_dest = OnlineWorkFrame1_1._frame_dest_reorganized(one_input_node)
        _data = np.asarray(data, dtype=PAYLOAD_DATA_DTYPE)

        return OnlineFrameGen.gen_work_frame1_1_fast(common_frame_dest, _data)

    @staticmethod
    def gen_work_frame1_1_fast(
        frame_dest_info: FrameArrayType, data: PayloadDataType
    ) -> FrameArrayType:
        if frame_dest_info.size != data.size:
            raise ValueError(
                f"the size of frame dest info & mask are not equal, {frame_dest_info.size} != {data.size}."
            )

        mask = np.flatnonzero(data)
        # Only where there is 1 returns.
        return frame_dest_info[mask]

    @staticmethod
    def gen_work_frame1_2(
        chip_coord: ChipCoord, coord: Coord, rid: RId = _rid_unset(), /
    ) -> OnlineWorkFrame1_2:
        return OnlineWorkFrame1_2(chip_coord, coord, rid)

    @staticmethod
    def gen_work_frame1_3(
        chip_coord: ChipCoord, coord: Coord, rid: RId = _rid_unset(), /
    ) -> OnlineWorkFrame1_3:
        return OnlineWorkFrame1_3(chip_coord, coord, rid)

    @staticmethod
    def gen_work_frame1_4(
        chip_coord: ChipCoord, coord: Coord, rid: RId = _rid_unset(), /
    ) -> OnlineWorkFrame1_4:
        return OnlineWorkFrame1_4(chip_coord, coord, rid)

    @staticmethod
    def gen_work_frame2(chip_coord: ChipCoord, /, n_sync: int) -> OnlineWorkFrame2:
        return OnlineWorkFrame2(chip_coord, n_sync)

    @staticmethod
    def gen_work_frame3(chip_coord: ChipCoord) -> OnlineWorkFrame3:
        return OnlineWorkFrame3(chip_coord)

    @staticmethod
    def gen_work_frame4(chip_coord: ChipCoord) -> OnlineWorkFrame4:
        return OnlineWorkFrame4(chip_coord)


class ChipFrameGen:
    fgen_handler: ClassVar[dict[CoreType, OfflineFrameGen | OnlineFrameGen]] = {
        CoreType.OFFLINE: OfflineFrameGen(),
        CoreType.ONLINE: OnlineFrameGen(),
    }

    @classmethod
    def gen_magic_init_frame(
        cls,
        chip_coord: ChipCoord,
        core_coord: Coord | Sequence[Coord],
        core_rid: RId | Sequence[RId] = _rid_unset(),
        redundant_init: bool = True,
    ) -> tuple[FrameArrayType, FrameArrayType]:
        """Magic initialization frames for PAICORE. DO NOT MODIFY!

        Args:
            chip_coord: coordinate of the target chip.
            core_coord: coordinates of the target cores.
            redundant_init: whether to use redundant initialization frames, in case of failure.

        If use redundant initialization frames, the magic frames are composed of:
            1. [config1[0] of core #1] + [init frame] + [config1[0] of core #2] + [init frame] + ...\
                + [config1[0] of core #N] + [init frame]

            2. [config1[1] of core #1] + [config1[2] of core #1] + [config1[1] of core #2] + ... +  \
                [config1[2] of core #2] + [config1[1] of core #N] + [config1[2] of core #N]
            3. [work1[0] of core #1] + [work1[0] of core #2] + ... + [work1[0] of core #N]

        Else,
            1. [config1[0] of core #1] + [config1[0] of core #2] + ... + [config1[0] of core #N] +  \
                [init frame]

            2, 3 remain the same.

        Returns: two parts of magic frames.
        """
        if isinstance(core_coord, Coord):
            _core_coord = (core_coord,)
        else:
            _core_coord = core_coord

        if isinstance(core_rid, RId):
            _core_rid = (core_rid,) * len(_core_coord)
        else:
            _core_rid = core_rid

        magic_frame_cf_1 = []
        magic_frame_cf_2 = []
        magic_frame_wf = []
        init_frame = OfflineWorkFrame4(chip_coord)

        for coord, rid in zip(_core_coord, _core_rid, strict=True):
            # NOTE: The length of config type I for online & offline cores are not the same.
            if coord.core_type == CoreType.OFFLINE:
                config1 = OfflineConfigFrame1(chip_coord, coord, rid, 0)
            else:
                config1 = OnlineConfigFrame1(chip_coord, coord, rid, None)

            work1 = OfflineWorkFrame1(chip_coord, coord, rid, 0, 0, 0)

            magic_frame_cf_1.append(config1.value[0])
            if redundant_init:
                magic_frame_cf_1.append(init_frame.value[0])

            magic_frame_cf_2.extend(config1.value[1:])
            magic_frame_wf.append(work1.value[0])

        if not redundant_init:
            magic_frame_cf_1.append(init_frame.value[0])

        magic_frame_cf_2.extend(magic_frame_wf)

        return np.asarray(magic_frame_cf_1, dtype=FRAME_DTYPE), np.asarray(
            magic_frame_cf_2, dtype=FRAME_DTYPE
        )



import warnings
from abc import ABC, abstractmethod
from typing import Any, ClassVar, Literal, overload

import numpy as np
from numpy.typing import ArrayLike, NDArray

from ..coordinate import ChipCoord, Coord, CoordLike, RIdLike, to_coord, to_rid
from ..coordinate import ReplicationId as RId
from ..core_defs import WeightWidth, core_mode_check
from ..core_model import CoreReg
from ..core_model import OfflineCoreReg as OffCoreReg
from ..core_model import OnlineCoreReg as OnCoreReg
from ..hw_defs import HwOfflineCoreParams as OffCoreParams
from ..hw_defs import HwOnlineCoreParams as OnCoreParams
from ..neuron_model import NeuAttrs, NeuDestInfo
from ..neuron_model import OfflineNeuAttrs as OffNeuAttrs
from ..neuron_model import OfflineNeuDestInfo as OffNeuDestInfo
from ..neuron_model import OfflineNeuDestInfoChecker as OffNeuDestInfoChecker
from ..neuron_model import OnlineNeuAttrs as OnNeuAttrs
from ..neuron_model import OnlineNeuDestInfo as OnNeuDestInfo
from ..neuron_model import OnlineNeuDestInfoChecker as OnNeuDestInfoChecker
from ..routing_defs import _rid_unset
from .base import Frame, FramePackage, FramePackagePayload, get_frame_dest
from .frame_defs import FrameFormat as FF
from .frame_defs import FrameHeader as FH
from .frame_defs import FramePackageType as FPType
from .frame_defs import OfflineCoreRegFormat as Off_CRegF
from .frame_defs import OfflineNeuRAMFormat as Off_NRAMF
from .frame_defs import OfflineWorkFrame1Format as Off_WF1F
from .frame_defs import OfflineWorkFrame2Format as Off_WF2F
from .frame_defs import Online_WF1F_SubType
from .frame_defs import OnlineCoreRegFormat as On_CRegF
from .frame_defs import OnlineNeuRAMFormat as On_NRAMF
from .frame_defs import OnlineNeuRAMFormat_WW1 as On_NRAMF_WW1
from .frame_defs import OnlineNeuRAMFormat_WWn as On_NRAMF_WWn
from .frame_defs import OnlineWorkFrame1Format as On_WF1F
from .frame_defs import OnlineWorkFrame1Format_1 as On_WF1F_1
from .types import (
    FRAME_DTYPE,
    LUT_DTYPE,
    PAYLOAD_DATA_DTYPE,
    DataType,
    FrameArrayType,
    IntScalarType,
    LUTDataType,
)
from .utils import (
    OUT_OF_RANGE_WARNING,
    ShapeError,
    TruncationWarning,
    _mask,
    bin_split,
    params_check,
)

__all__ = [
    "OfflineConfigFrame1",
    "OfflineConfigFrame2",
    "OfflineConfigFrame3",
    "OfflineConfigFrame4",
    "OfflineTestInFrame1",
    "OfflineTestInFrame2",
    "OfflineTestInFrame3",
    "OfflineTestInFrame4",
    "OfflineTestOutFrame1",
    "OfflineTestOutFrame2",
    "OfflineTestOutFrame3",
    "OfflineTestOutFrame4",
    "OfflineWorkFrame1",
    "OfflineWorkFrame2",
    "OfflineWorkFrame3",
    "OfflineWorkFrame4",
    "OnlineConfigFrame1",
    "OnlineConfigFrame2",
    "OnlineConfigFrame3",
    "OnlineConfigFrame4",
    "OnlineTestInFrame1",
    "OnlineTestInFrame2",
    "OnlineTestInFrame3",
    "OnlineTestInFrame4",
    "OnlineTestOutFrame1",
    "OnlineTestOutFrame2",
    "OnlineTestOutFrame3",
    "OnlineTestOutFrame4",
    "OnlineWorkFrame1_1",
    "OnlineWorkFrame1_2",
    "OnlineWorkFrame1_3",
    "OnlineWorkFrame1_4",
    "OnlineWorkFrame2",
    "OnlineWorkFrame3",
    "OnlineWorkFrame4",
]


class _RandomSeedFrame(Frame):
    N_FRAME_PAYLOAD: ClassVar[int] = 3

    def __init__(
        self, chip_coord: ChipCoord, core_coord: Coord, rid: RId, random_seed: int
    ) -> None:
        if random_seed > FF.GENERAL_MASK:
            warnings.warn(
                OUT_OF_RANGE_WARNING.format(
                    "random_seed", FF.FRAME_LENGTH, random_seed
                ),
                TruncationWarning,
            )

        payload = self._random_seed_split(random_seed)
        super().__init__(chip_coord, core_coord, rid, payload)

    @staticmethod
    def _random_seed_split(random_seed: int) -> FrameArrayType:
        _seed = random_seed & FF.GENERAL_MASK

        return np.asarray(
            [
                (_seed >> 34) & FF.GENERAL_PAYLOAD_MASK,
                (_seed >> 4) & FF.GENERAL_PAYLOAD_MASK,
                (_seed & _mask(4)) << (FF.GENERAL_PAYLOAD_LENGTH - 4),
            ],
            dtype=FRAME_DTYPE,
        )


class _CoreRegFrame(Frame):
    N_FRAME_PAYLOAD: ClassVar[int]

    def __init__(
        self, chip_coord: ChipCoord, core_coord: Coord, rid: RId, core_reg: CoreReg
    ) -> None:
        core_reg_dict = core_reg.model_dump()
        payload = self.payload_reorganized(core_reg_dict)
        super().__init__(chip_coord, core_coord, rid, payload)

    @staticmethod
    @abstractmethod
    def payload_reorganized(core_reg: dict[str, Any]) -> FrameArrayType: ...

    @property
    def params_reg(self) -> FRAME_DTYPE | FrameArrayType:
        return self.payload


@overload
def _convert_core_reg(
    core_reg: OffCoreReg | dict[str, Any], type: Literal["offline"]
) -> OffCoreReg: ...


@overload
def _convert_core_reg(
    core_reg: OnCoreReg | dict[str, Any], type: Literal["online"]
) -> OnCoreReg: ...


def _convert_core_reg(
    core_reg: CoreReg | dict[str, Any], type: Literal["offline", "online"]
) -> OffCoreReg | OnCoreReg:
    if isinstance(core_reg, (OffCoreReg, OnCoreReg)):
        return core_reg
    else:
        if type == "offline":
            return OffCoreReg.model_validate(core_reg, strict=True)
        else:
            return OnCoreReg.model_validate(core_reg, strict=True)


class _OfflineCoreRegFrame(_CoreRegFrame):
    N_FRAME_PAYLOAD = 3

    def __init__(
        self,
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        core_reg: OffCoreReg | dict[str, Any],
    ) -> None:
        _core_reg = _convert_core_reg(core_reg, "offline")
        super().__init__(chip_coord, core_coord, rid, _core_reg)

    @staticmethod
    @core_mode_check
    def payload_reorganized(core_reg: dict[str, Any]) -> FrameArrayType:
        # High 8 bits & low 7 bits of 'tick_wait_start'
        tws_high8, tws_low7 = bin_split(core_reg["tick_wait_start"], 7, 8)
        reg_frame1 = (
            (core_reg["weight_width"] & Off_CRegF.WEIGHT_WIDTH_MASK)
            << Off_CRegF.WEIGHT_WIDTH_OFFSET
            | ((core_reg["lcn"] & Off_CRegF.LCN_MASK) << Off_CRegF.LCN_OFFSET)
            | (
                (core_reg["input_width"] & Off_CRegF.INPUT_WIDTH_MASK)
                << Off_CRegF.INPUT_WIDTH_OFFSET
            )
            | (
                (core_reg["spike_width"] & Off_CRegF.SPIKE_WIDTH_MASK)
                << Off_CRegF.SPIKE_WIDTH_OFFSET
            )
            | (
                (core_reg["num_dendrite"] & Off_CRegF.NUM_DENDRITE_MASK)
                << Off_CRegF.NUM_DENDRITE_OFFSET
            )
            | (
                (core_reg["max_pooling_en"] & Off_CRegF.MAX_POOLING_EN_MASK)
                << Off_CRegF.MAX_POOLING_EN_OFFSET
            )
            | (
                (tws_high8 & Off_CRegF.TICK_WAIT_START_HIGH8_MASK)
                << Off_CRegF.TICK_WAIT_START_HIGH8_OFFSET
            )
        )

        # High 3 bits & low 7 bits of 'test_chip_addr'
        tca_high3, tca_low7 = bin_split(core_reg["test_chip_addr"], 7, 3)
        reg_frame2 = (
            (
                (tws_low7 & Off_CRegF.TICK_WAIT_START_LOW7_MASK)
                << Off_CRegF.TICK_WAIT_START_LOW7_OFFSET
            )
            | (
                (core_reg["tick_wait_end"] & Off_CRegF.TICK_WAIT_END_MASK)
                << Off_CRegF.TICK_WAIT_END_OFFSET
            )
            | ((core_reg["snn_en"] & Off_CRegF.SNN_EN_MASK) << Off_CRegF.SNN_EN_OFFSET)
            | (
                (core_reg["target_lcn"] & Off_CRegF.TARGET_LCN_MASK)
                << Off_CRegF.TARGET_LCN_OFFSET
            )
            | (
                (tca_high3 & Off_CRegF.TEST_CHIP_ADDR_HIGH3_MASK)
                << Off_CRegF.TEST_CHIP_ADDR_HIGH3_OFFSET
            )
        )
        reg_frame3 = (
            tca_low7 & Off_CRegF.TEST_CHIP_ADDR_LOW7_MASK
        ) << Off_CRegF.TEST_CHIP_ADDR_LOW7_OFFSET

        return np.asarray([reg_frame1, reg_frame2, reg_frame3], dtype=FRAME_DTYPE)


class _OnlineCoreRegFrame(_CoreRegFrame):
    N_FRAME_PAYLOAD = 8

    def __init__(
        self,
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        core_reg: OnCoreReg | dict[str, Any],
    ) -> None:
        _core_reg = _convert_core_reg(core_reg, "online")
        super().__init__(chip_coord, core_coord, rid, _core_reg)

    @staticmethod
    def payload_reorganized(core_reg: dict[str, Any]) -> FrameArrayType:
        # High 26 bits & low 6 bits of 'lateral_inhi_value'
        lateral_inhi_high26, lateral_inhi_low6 = bin_split(
            core_reg["lateral_inhi_value"], 6, 26
        )
        reg_frame1 = (
            (core_reg["weight_width"] & On_CRegF.WEIGHT_WIDTH_MASK)
            << On_CRegF.WEIGHT_WIDTH_OFFSET
            | ((core_reg["lcn"] & On_CRegF.LCN_MASK) << On_CRegF.LCN_OFFSET)
            | (
                (lateral_inhi_high26 & On_CRegF.LATERAL_INHI_VALUE_HIGH26_MASK)
                << On_CRegF.LATERAL_INHI_VALUE_HIGH26_OFFSET
            )
        )

        reg_frame2 = (
            (lateral_inhi_low6 & On_CRegF.LATERAL_INHI_VALUE_LOW6_MASK)
            << On_CRegF.LATERAL_INHI_VALUE_LOW6_OFFSET
            | (
                (core_reg["weight_decay_value"] & On_CRegF.WEIGHT_DECAY_MASK)
                << On_CRegF.WEIGHT_DECAY_OFFSET
            )
            | (
                (core_reg["upper_weight"] & On_CRegF.UPPER_WEIGHT_MASK)
                << On_CRegF.UPPER_WEIGHT_OFFSET
            )
            | (
                (core_reg["lower_weight"] & On_CRegF.LOWER_WEIGHT_MASK)
                << On_CRegF.LOWER_WEIGHT_OFFSET
            )
        )

        reg_frame3 = (
            (
                (core_reg["neuron_start"] & On_CRegF.NEURON_START_MASK)
                << On_CRegF.NEURON_START_OFFSET
            )
            | (
                (core_reg["neuron_end"] & On_CRegF.NEURON_END_MASK)
                << On_CRegF.NEURON_END_OFFSET
            )
            | (
                (core_reg["inhi_core_x_ex"] & On_CRegF.INHI_CORE_X_EX_MASK)
                << On_CRegF.INHI_CORE_X_EX_OFFSET
            )
            | (
                (core_reg["inhi_core_y_ex"] & On_CRegF.INHI_CORE_Y_EX_MASK)
                << On_CRegF.INHI_CORE_Y_EX_OFFSET
            )
        )

        reg_frame4 = (
            (core_reg["tick_wait_start"] & On_CRegF.TICK_WAIT_START_MASK)
            << On_CRegF.TICK_WAIT_START_OFFSET
        ) | (
            (core_reg["tick_wait_end"] & On_CRegF.TICK_WAIT_END_MASK)
            << On_CRegF.TICK_WAIT_END_OFFSET
        )

        # High 30 bits & low 30 bits of 'lut_random_en'
        lut_random_en_high30, lut_random_en_low30 = bin_split(
            core_reg["lut_random_en"], 30, 30
        )
        reg_frame5 = (
            lut_random_en_high30 & On_CRegF.LUT_RANDOM_EN_HIGH30_MASK
        ) << On_CRegF.LUT_RANDOM_EN_HIGH30_OFFSET
        reg_frame6 = (
            lut_random_en_low30 & On_CRegF.LUT_RANDOM_EN_LOW30_MASK
        ) << On_CRegF.LUT_RANDOM_EN_LOW30_OFFSET

        reg_frame7 = (
            (
                (core_reg["decay_random_en"] & On_CRegF.DECAY_RANDOM_EN_MASK)
                << On_CRegF.DECAY_RANDOM_EN_MASK
            )
            | (
                (core_reg["leak_order"] & On_CRegF.LEAK_ORDER_MASK)
                << On_CRegF.LEAK_ORDER_OFFSET
            )
            | (
                (core_reg["online_mode_en"] & On_CRegF.ONLINE_MODE_EN_MASK)
                << On_CRegF.ONLINE_MODE_EN_OFFSET
            )
            | (
                (core_reg["test_chip_addr"] & On_CRegF.TEST_CHIP_ADDR_MASK)
                << On_CRegF.TEST_CHIP_ADDR_OFFSET
            )
            | (
                (core_reg["random_seed"] & On_CRegF.RANDOM_SEED_MASK)
                << On_CRegF.RANDOM_SEED_OFFSET
            )
        )
        reg_frame8 = 0  # 30'd0

        return np.asarray(
            [
                reg_frame1,
                reg_frame2,
                reg_frame3,
                reg_frame4,
                reg_frame5,
                reg_frame6,
                reg_frame7,
                reg_frame8,
            ],
            dtype=FRAME_DTYPE,
        )


class _NeuRAMFrame(FramePackage, ABC):
    N_FRAME_PAYLOAD: int

    def __init__(
        self,
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        ram_start_addr: int,
        n_neuron: int,
        neu_attrs: NeuAttrs,
        neu_dest_info: NeuDestInfo,
        repeat: int,
    ) -> None:
        n_package = self.N_FRAME_PAYLOAD * n_neuron * repeat
        payload = FramePackagePayload(ram_start_addr, FPType.CONF_TESTOUT, n_package)
        neu_attrs_dict = neu_attrs.model_dump()
        neu_dest_info_dict = neu_dest_info.model_dump()
        packages = self.get_packages(
            neu_attrs_dict, neu_dest_info_dict, n_neuron, repeat
        )

        super().__init__(chip_coord, core_coord, rid, payload, packages)

    @abstractmethod
    def get_packages(
        self,
        attrs: dict[str, Any],
        dest_info: dict[str, Any],
        n_neuron: int,
        repeat: int,
    ) -> FrameArrayType: ...


@overload
def _convert_neu_attrs_dest_info(
    neu_attrs: OffNeuAttrs | dict[str, Any],
    neu_dest_info: OffNeuDestInfo | dict[str, Any],
    type: Literal["offline"],
) -> tuple[OffNeuAttrs, OffNeuDestInfo]: ...


@overload
def _convert_neu_attrs_dest_info(
    neu_attrs: OnNeuAttrs | dict[str, Any],
    neu_dest_info: OnNeuDestInfo | dict[str, Any],
    type: Literal["online"],
    ww: WeightWidth,
) -> tuple[OnNeuAttrs, OnNeuDestInfo]: ...


def _convert_neu_attrs_dest_info(
    neu_attrs: NeuAttrs | dict[str, Any],
    neu_dest_info: NeuDestInfo | dict[str, Any],
    type: Literal["offline", "online"],
    ww: WeightWidth | None = None,
):
    if isinstance(neu_attrs, (OffNeuAttrs, OnNeuAttrs)):
        _neu_attrs = neu_attrs
    else:
        if type == "offline":
            _neu_attrs = OffNeuAttrs.model_validate(neu_attrs, strict=True)
        else:
            _neu_attrs = OnNeuAttrs.model_validate(
                neu_attrs, context={"weight_width": ww}, strict=True
            )

    if isinstance(neu_dest_info, (OffNeuDestInfo, OnNeuDestInfo)):
        _neu_dest_info = neu_dest_info
    else:
        if type == "offline":
            _neu_dest_info = OffNeuDestInfo.model_validate(neu_dest_info, strict=True)
        else:
            _neu_dest_info = OnNeuDestInfo.model_validate(neu_dest_info, strict=True)

    return _neu_attrs, _neu_dest_info


class _OfflineNeuRAMFrame(_NeuRAMFrame):
    N_FRAME_PAYLOAD = 4  # One neuron needs 4 frames to configure

    def __init__(
        self,
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        ram_start_addr: int,
        n_neuron: int,
        neu_attrs: OffNeuAttrs | dict[str, Any],
        neu_dest_info: OffNeuDestInfo | dict[str, Any],
        repeat: int,
    ) -> None:
        _neu_attrs, _neu_dest_info = _convert_neu_attrs_dest_info(
            neu_attrs, neu_dest_info, "offline"
        )

        super().__init__(
            chip_coord,
            core_coord,
            rid,
            ram_start_addr,
            n_neuron,
            _neu_attrs,
            _neu_dest_info,
            repeat,
        )

    def get_packages(
        self,
        attrs: dict[str, Any],
        dest_info: dict[str, Any],
        n_neuron: int,
        repeat: int,
    ) -> FrameArrayType:
        voltage: int = attrs.get("voltage", 0)

        def _gen_ram_frame1_and_2(leak_v: int) -> tuple[FRAME_DTYPE, FRAME_DTYPE]:
            leak_v_high2, leak_v_low28 = bin_split(leak_v, 28, 2)

            # Package #1, [63:0]
            rf1 = (
                ((voltage & Off_NRAMF.VOLTAGE_MASK) << Off_NRAMF.VOLTAGE_OFFSET)
                | (
                    (attrs["bit_trunc"] & Off_NRAMF.BIT_TRUNC_MASK)
                    << Off_NRAMF.BIT_TRUNC_OFFSET
                )
                | (
                    (
                        attrs["syn_integration_mode"]
                        & Off_NRAMF.SYN_INTEGRATION_MODE_MASK
                    )
                    << Off_NRAMF.SYN_INTEGRATION_MODE_OFFSET
                )
                | (
                    (leak_v_low28 & Off_NRAMF.LEAK_V_LOW28_MASK)
                    << Off_NRAMF.LEAK_V_LOW28_OFFSET
                )
            )
            # Package #2, [127:64]
            rf2 = (
                (
                    (leak_v_high2 & Off_NRAMF.LEAK_V_HIGH2_MASK)
                    << Off_NRAMF.LEAK_V_HIGH2_OFFSET
                )
                | (
                    (
                        attrs["leak_integration_mode"]
                        & Off_NRAMF.LEAK_INTEGRATION_MODE_MASK
                    )
                    << Off_NRAMF.LEAK_INTEGRATION_MODE_OFFSET
                )
                | (
                    (attrs["leak_direction"] & Off_NRAMF.LEAK_DIRECTION_MASK)
                    << Off_NRAMF.LEAK_DIRECTION_OFFSET
                )
                | (
                    (attrs["pos_threshold"] & Off_NRAMF.POS_THRESHOLD_MASK)
                    << Off_NRAMF.POS_THRESHOLD_OFFSET
                )
                | (
                    (attrs["neg_threshold"] & Off_NRAMF.NEG_THRES_MASK)
                    << Off_NRAMF.NEG_THRES_OFFSET
                )
                | (
                    (attrs["neg_thres_mode"] & Off_NRAMF.NEG_THRES_MODE_MASK)
                    << Off_NRAMF.NEG_THRES_MODE_OFFSET
                )
                | (
                    (thres_mask_bit_low1 & Off_NRAMF.THRE_MASK_BITS_LOW1_MASK)
                    << Off_NRAMF.THRE_MASK_BITS_LOW1_OFFSET
                )
            )

            return FRAME_DTYPE(rf1), FRAME_DTYPE(rf2)

        def _gen_ram_frame3() -> FRAME_DTYPE:
            # Package #3, [191:128]
            return FRAME_DTYPE(
                (
                    (thres_mask_bit_high4 & Off_NRAMF.THRES_MASK_BITS_HIGH4_MASK)
                    << Off_NRAMF.THRES_MASK_BITS_HIGH4_OFFSET
                )
                | (
                    (attrs["leak_comparison"] & Off_NRAMF.LEAK_COMPARISON_MASK)
                    << Off_NRAMF.LEAK_COMPARISON_OFFSET
                )
                | (
                    (attrs["reset_v"] & Off_NRAMF.RESET_V_MASK)
                    << Off_NRAMF.RESET_V_OFFSET
                )
                | (
                    (attrs["reset_mode"] & Off_NRAMF.RESET_MODE_MASK)
                    << Off_NRAMF.RESET_MODE_OFFSET
                )
                | (
                    (dest_info["addr_chip_y"] & Off_NRAMF.ADDR_CHIP_Y_MASK)
                    << Off_NRAMF.ADDR_CHIP_Y_OFFSET
                )
                | (
                    (dest_info["addr_chip_x"] & Off_NRAMF.ADDR_CHIP_X_MASK)
                    << Off_NRAMF.ADDR_CHIP_X_OFFSET
                )
                | (
                    (dest_info["addr_core_y_ex"] & Off_NRAMF.ADDR_CORE_Y_EX_MASK)
                    << Off_NRAMF.ADDR_CORE_Y_EX_OFFSET
                )
                | (
                    (dest_info["addr_core_x_ex"] & Off_NRAMF.ADDR_CORE_X_EX_MASK)
                    << Off_NRAMF.ADDR_CORE_X_EX_OFFSET
                )
                | (
                    (dest_info["addr_core_y"] & Off_NRAMF.ADDR_CORE_Y_MASK)
                    << Off_NRAMF.ADDR_CORE_Y_OFFSET
                )
                | (
                    (addr_core_x_low2 & Off_NRAMF.ADDR_CORE_X_LOW2_MASK)
                    << Off_NRAMF.ADDR_CORE_X_LOW2_OFFSET
                )
            )

        def _gen_ram_frame4(idx: int) -> FRAME_DTYPE:
            # Package #4, [213:192]
            return FRAME_DTYPE(
                (
                    (addr_core_x_high3 & Off_NRAMF.ADDR_CORE_X_HIGH3_MASK)
                    << Off_NRAMF.ADDR_CORE_X_HIGH3_OFFSET
                )
                | (
                    (addr_axon[idx] & Off_NRAMF.ADDR_AXON_MASK)
                    << Off_NRAMF.ADDR_AXON_OFFSET
                )
                | (
                    (timeslot[idx] & Off_NRAMF.TICK_RELATIVE_MASK)
                    << Off_NRAMF.TICK_RELATIVE_OFFSET
                )
            )

        timeslot = dest_info["tick_relative"]
        addr_axon = dest_info["addr_axon"]

        if len(timeslot) != len(addr_axon):
            raise ValueError(
                f"length of 'tick_relative' & 'addr_axon' are not equal, "
                f"{len(timeslot)} != {len(addr_axon)}"
            )

        if n_neuron > len(timeslot):
            raise ValueError(f"length of 'tick_relative' out of range ({n_neuron})")

        packages = np.zeros((n_neuron, 4), dtype=FRAME_DTYPE)
        thres_mask_bit_high4, thres_mask_bit_low1 = bin_split(
            attrs["thres_mask_bits"], 1, 4
        )
        addr_core_x_high3, addr_core_x_low2 = bin_split(dest_info["addr_core_x"], 2, 3)

        # LSB: [63:0], [127:64], [191:128], [213:192]
        rf3 = _gen_ram_frame3()

        leak_v: int | NDArray[np.int32] = attrs["leak_v"]
        if isinstance(leak_v, np.ndarray):
            if leak_v.size != n_neuron:
                raise ValueError(
                    f"length of 'leak_v' is not equal to #N of neuron, "
                    f"{leak_v.size} != {n_neuron}"
                )
        else:
            leak_v = np.full(n_neuron, leak_v, dtype=np.int32)

        # Vectorize the functions
        gen_ram_frame1_and_2 = np.vectorize(
            _gen_ram_frame1_and_2, otypes=[FRAME_DTYPE, FRAME_DTYPE]
        )
        gen_ram_frame4 = np.vectorize(_gen_ram_frame4, otypes=[FRAME_DTYPE])

        rf1, rf2 = gen_ram_frame1_and_2(leak_v)
        rf4 = gen_ram_frame4(np.arange(n_neuron))

        packages[:, 0] = rf1
        packages[:, 1] = rf2
        packages[:, 2] = rf3
        packages[:, 3] = rf4

        # Tile the package of every neuron `repeat` times & flatten
        # (n_neuron, 4) -> (n_neuron * 4 * repeat,)
        packages_tiled = np.tile(packages, repeat).ravel()

        return packages_tiled


class _OnlineNeuRAMFrame(_NeuRAMFrame):
    def __init__(
        self,
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        ram_start_addr: int,
        n_neuron: int,
        neu_attrs: OnNeuAttrs | dict[str, Any],
        neu_dest_info: OnNeuDestInfo | dict[str, Any],
        weight_width: WeightWidth,
    ) -> None:
        # NOTE: When weight width is 1bit, a neuron need 2 frames (1 address) to store attributes.
        # Otherwise, a neuron need 4 frames (2 addresses) to store attributes.
        if weight_width == WeightWidth.WEIGHT_WIDTH_1BIT:
            self.N_FRAME_PAYLOAD = 2
        else:
            self.N_FRAME_PAYLOAD = 4

        _neu_attrs, _neu_dest_info = _convert_neu_attrs_dest_info(
            neu_attrs, neu_dest_info, "online", weight_width
        )

        super().__init__(
            chip_coord,
            core_coord,
            rid,
            ram_start_addr,
            n_neuron,
            _neu_attrs,
            _neu_dest_info,
            1,
        )

    def get_packages(
        self,
        attrs: dict[str, Any],
        dest_info: dict[str, Any],
        n_neuron: int,
        repeat: int,
    ) -> FrameArrayType:
        def _gen_ram_frame1(idx: int) -> FRAME_DTYPE:
            # Package #1, [63:0]
            return FRAME_DTYPE(
                (
                    (attrs["plasticity_end"] & On_NRAMF.PLASTICITY_END_MASK)
                    << On_NRAMF.PLASTICITY_END_OFFSET
                )
                | (
                    (attrs["plasticity_start"] & On_NRAMF.PLASTICITY_START_MASK)
                    << On_NRAMF.PLASTICITY_START_OFFSET
                )
                | (
                    (addr_axon[idx] & On_NRAMF.ADDR_AXON_MASK)
                    << On_NRAMF.ADDR_AXON_OFFSET
                )
                | (
                    (dest_info["addr_core_y_ex"] & On_NRAMF.ADDR_CORE_Y_EX_MASK)
                    << On_NRAMF.ADDR_CORE_Y_EX_OFFSET
                )
                | (
                    (dest_info["addr_core_x_ex"] & On_NRAMF.ADDR_CORE_X_EX_MASK)
                    << On_NRAMF.ADDR_CORE_X_EX_OFFSET
                )
                | (
                    (dest_info["addr_core_y"] & On_NRAMF.ADDR_CORE_Y_MASK)
                    << On_NRAMF.ADDR_CORE_Y_OFFSET
                )
                | (
                    (dest_info["addr_core_x"] & On_NRAMF.ADDR_CORE_X_MASK)
                    << On_NRAMF.ADDR_CORE_X_OFFSET
                )
                | (
                    (timeslot[idx] & On_NRAMF.TIME_RELATIVE_MASK)
                    << On_NRAMF.TIME_RELATIVE_OFFSET
                )
            )

        def _gen_frame2(leak_v: int, init_v: int) -> FRAME_DTYPE:
            # Package #2, [127:64]
            return FRAME_DTYPE(
                (
                    (attrs["voltage"] & On_NRAMF_WW1.VOLTAGE_MASK)
                    << On_NRAMF_WW1.VOLTAGE_OFFSET
                )
                | ((init_v & On_NRAMF_WW1.INIT_V_MASK) << On_NRAMF_WW1.INIT_V_OFFSET)
                | (
                    (attrs["reset_v"] & On_NRAMF_WW1.RESET_V_MASK)
                    << On_NRAMF_WW1.RESET_V_OFFSET
                )
                | (
                    (attrs["neg_threshold"] & On_NRAMF_WW1.NEG_THRES_MASK)
                    << On_NRAMF_WW1.NEG_THRES_OFFSET
                )
                | (
                    (attrs["pos_threshold"] & On_NRAMF_WW1.POS_THRES_MASK)
                    << On_NRAMF_WW1.POS_THRES_OFFSET
                )
                | ((leak_v & On_NRAMF_WW1.LEAK_V_MASK) << On_NRAMF_WW1.LEAK_V_OFFSET)
            )

        def _gen_frame_2to4(
            leak_v: FRAME_DTYPE, init_v: FRAME_DTYPE
        ) -> tuple[FRAME_DTYPE, FRAME_DTYPE, FRAME_DTYPE]:
            # Package #2, [127:64]
            rf2 = (
                (attrs["voltage"] & On_NRAMF_WWn.VOLTAGE_MASK)
                << On_NRAMF_WWn.VOLTAGE_OFFSET
            ) | ((init_v & On_NRAMF_WWn.INIT_V_MASK) << On_NRAMF_WWn.INIT_V_OFFSET)
            # Package #3, [191:128]
            rf3 = (
                (attrs["reset_v"] & On_NRAMF_WWn.RESET_V_MASK)
                << On_NRAMF_WWn.RESET_V_OFFSET
            ) | (
                (attrs["neg_threshold"] & On_NRAMF_WWn.NEG_THRES_MASK)
                << On_NRAMF_WWn.NEG_THRES_OFFSET
            )
            # Package #4, [255:192]
            rf4 = (
                (attrs["pos_threshold"] & On_NRAMF_WWn.POS_THRES_MASK)
                << On_NRAMF_WWn.POS_THRES_OFFSET
            ) | ((leak_v & On_NRAMF_WWn.LEAK_V_MASK) << On_NRAMF_WWn.LEAK_V_OFFSET)

            return (FRAME_DTYPE(rf2), FRAME_DTYPE(rf3), FRAME_DTYPE(rf4))

        timeslot = dest_info["tick_relative"]
        addr_axon = dest_info["addr_axon"]

        if len(timeslot) != len(addr_axon):
            raise ValueError(
                f"length of 'tick_relative' & 'addr_axon' are not equal, "
                f"{len(timeslot)} != {len(addr_axon)}"
            )

        if n_neuron > len(timeslot):
            raise ValueError(f"length of 'tick_relative' out of range ({n_neuron})")

        packages = np.zeros((n_neuron, self.N_FRAME_PAYLOAD), dtype=FRAME_DTYPE)

        leak_v: int | NDArray[np.int32] = attrs["leak_v"]
        init_v: int | NDArray[np.int32] = attrs["init_v"]

        if isinstance(leak_v, np.ndarray):
            if leak_v.size != n_neuron:
                raise ValueError(
                    f"length of 'leak_v' is not equal to #N of neuron, "
                    f"{leak_v.size} != {n_neuron}"
                )
        else:
            leak_v = np.full(n_neuron, leak_v, dtype=np.int32)

        if isinstance(init_v, np.ndarray):
            if init_v.size != n_neuron:
                raise ValueError(
                    f"length of 'init_v' is not equal to #N of neuron, "
                    f"{init_v.size} != {n_neuron}"
                )
        else:
            init_v = np.full(n_neuron, init_v, dtype=np.int32)

        # Vectorize the functions
        gen_ram_frame1 = np.vectorize(_gen_ram_frame1, otypes=[FRAME_DTYPE])
        gen_frame2 = np.vectorize(_gen_frame2, otypes=[FRAME_DTYPE])
        gen_frame_2to4 = np.vectorize(
            _gen_frame_2to4, otypes=[FRAME_DTYPE, FRAME_DTYPE, FRAME_DTYPE]
        )

        rf1 = gen_ram_frame1(np.arange(n_neuron))

        if self.N_FRAME_PAYLOAD == 2:
            rf_after1 = gen_frame2(leak_v, init_v).reshape(-1, 1)
        else:
            rf_after1 = np.column_stack(gen_frame_2to4(leak_v, init_v))

        packages[:, 0] = rf1
        packages[:, 1:] = rf_after1

        return packages.ravel()


class _WeightRAMFrame(FramePackage, ABC):
    N_WRAM_ADDR: ClassVar[int]
    N_FRAME_PER_WRAM: ClassVar[int]
    N_FRAME_PER_NRAM: ClassVar[int]

    def __init__(
        self,
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        ram_start_addr: int,
        n_package: int,
        weight_ram: FrameArrayType,
    ) -> None:
        if weight_ram.size != n_package:
            raise ValueError(
                f"size of weigh ram must be the #N of packages ({n_package}), but got {weight_ram.size}"
            )

        payload = FramePackagePayload(ram_start_addr, FPType.CONF_TESTOUT, n_package)
        _weight_ram = weight_ram.ravel()

        super().__init__(chip_coord, core_coord, rid, payload, _weight_ram)


class _OfflineWeightRAMFrame(_WeightRAMFrame):
    # Total WRAM: 576Kb
    N_WRAM_ADDR = 512
    N_FRAME_PER_WRAM = 18  # 152 bits per wram address
    N_FRAME_PER_NRAM = (OffCoreParams.ADDR_AXON_MAX + 1) // FF.GENERAL_PACKAGE_LEN


class _OnlineWeightRAMFrame(_WeightRAMFrame):
    # Total WRAM: 1024Kb
    N_WRAM_ADDR = 8192
    N_FRAME_PER_WRAM = 2  # 128 bits per wram address
    N_FRAME_PER_NRAM = (OnCoreParams.ADDR_AXON_MAX + 1) // FF.GENERAL_PACKAGE_LEN


class _LUTRAMFrame(Frame):
    N_FRAME_PER_LUT_RAM: ClassVar[int] = 16

    def __init__(
        self,
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        lut: LUTDataType | None = None,
    ) -> None:
        if lut is None:
            lut = np.zeros((OnCoreParams.LUT_LEN,), dtype=LUT_DTYPE)

        if lut.size != OnCoreParams.LUT_LEN:
            raise ValueError(
                f"size of lut must be {OnCoreParams.LUT_LEN}, but got {lut.size}"
            )

        payload = self.payload_reorganized(lut)
        super().__init__(chip_coord, core_coord, rid, payload)

    @staticmethod
    def payload_reorganized(lut: LUTDataType) -> FrameArrayType:
        lut_u8 = lut.ravel().view(np.uint8).astype(FRAME_DTYPE)
        lut_payload = np.zeros((_LUTRAMFrame.N_FRAME_PER_LUT_RAM,), dtype=FRAME_DTYPE)

        # 15 LUTs per group, 4 groups, 16 frames
        for i in range(4):
            lut_payload[4 * i] = (
                (lut_u8[15 * i] << FRAME_DTYPE(22))
                + (lut_u8[15 * i + 1] << FRAME_DTYPE(14))
                + (lut_u8[15 * i + 2] << FRAME_DTYPE(6))
                + (lut_u8[15 * i + 3] >> FRAME_DTYPE(2))  # high6
            )
            lut_payload[4 * i + 1] = (
                ((lut_u8[15 * i + 3] & FRAME_DTYPE(0x03)) << FRAME_DTYPE(28))  # low2
                + (lut_u8[15 * i + 4] << FRAME_DTYPE(20))
                + (lut_u8[15 * i + 5] << FRAME_DTYPE(12))
                + (lut_u8[15 * i + 6] << FRAME_DTYPE(4))
                + (lut_u8[15 * i + 7] >> FRAME_DTYPE(4))  # high4
            )
            lut_payload[4 * i + 2] = (
                ((lut_u8[15 * i + 7] & FRAME_DTYPE(0x0F)) << FRAME_DTYPE(26))  # low4
                + (lut_u8[15 * i + 8] << FRAME_DTYPE(18))
                + (lut_u8[15 * i + 9] << FRAME_DTYPE(10))
                + (lut_u8[15 * i + 10] << FRAME_DTYPE(2))
                + (lut_u8[15 * i + 11] >> FRAME_DTYPE(6))  # low2
            )
            lut_payload[4 * i + 3] = (
                ((lut_u8[15 * i + 11] & FRAME_DTYPE(0x3F)) << FRAME_DTYPE(24))  # low6
                + (lut_u8[15 * i + 12] << FRAME_DTYPE(16))
                + (lut_u8[15 * i + 13] << FRAME_DTYPE(8))
                + lut_u8[15 * i + 14]
            )

        return lut_payload


"""Offline Frames"""


class OfflineConfigFrame1(_RandomSeedFrame):
    header = FH.CONFIG_TYPE1


class OfflineConfigFrame2(_OfflineCoreRegFrame):
    header = FH.CONFIG_TYPE2


class OfflineConfigFrame3(_OfflineNeuRAMFrame):
    header = FH.CONFIG_TYPE3


class OfflineConfigFrame4(_OfflineWeightRAMFrame):
    header = FH.CONFIG_TYPE4


class OfflineTestInFrame1(Frame):
    header = FH.TEST_TYPE1

    def __init__(self, chip_coord: ChipCoord, core_coord: Coord, rid: RId) -> None:
        super().__init__(chip_coord, core_coord, rid)


class OfflineTestOutFrame1(_RandomSeedFrame):
    header = FH.TEST_TYPE1


class OfflineTestInFrame2(Frame):
    header = FH.TEST_TYPE2

    def __init__(self, chip_coord: ChipCoord, core_coord: Coord, rid: RId) -> None:
        super().__init__(chip_coord, core_coord, rid)


class OfflineTestOutFrame2(_OfflineCoreRegFrame):
    header = FH.TEST_TYPE2


class OfflineTestInFrame3(FramePackage):
    header = FH.TEST_TYPE3

    def __init__(
        self,
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        ram_start_addr: int,
        n_package: int,
    ) -> None:
        payload = FramePackagePayload(ram_start_addr, FPType.TESTIN, n_package)
        super().__init__(chip_coord, core_coord, rid, payload)


class OfflineTestOutFrame3(_OfflineNeuRAMFrame):
    header = FH.TEST_TYPE3


class OfflineTestInFrame4(FramePackage):
    header = FH.TEST_TYPE4

    def __init__(
        self,
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        ram_start_addr: int,
        n_package: int,
    ) -> None:
        payload = FramePackagePayload(ram_start_addr, FPType.TESTIN, n_package)
        super().__init__(chip_coord, core_coord, rid, payload)


class OfflineTestOutFrame4(_OfflineWeightRAMFrame):
    header = FH.TEST_TYPE4


class OfflineWorkFrame1(Frame):
    header = FH.WORK_TYPE1

    def __init__(
        self,
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        timeslot: IntScalarType,
        axon: IntScalarType,
        data: DataType = 0,  # signed int8
    ) -> None:
        self.ax = int(axon)
        self.ts = int(timeslot)

        if self.ts > Off_WF1F.TIMESLOT_MASK or self.ts < 0:
            raise ValueError(
                f"timeslot out of range {Off_WF1F.TIMESLOT_MASK} ({self.ts})."
            )

        if self.ax > OffCoreParams.ADDR_AXON_MAX or self.ax < 0:
            raise ValueError(
                f"axon out of range {OffCoreParams.ADDR_AXON_MAX} ({self.ax})."
            )

        if isinstance(data, np.ndarray) and data.size != 1:
            raise ShapeError(f"size of data must be 1 ({data.size}).")

        if data < np.iinfo(np.int8).min or data > np.iinfo(np.int8).max:
            raise ValueError(f"data out of range np.int8 ({data}).")

        self.data = PAYLOAD_DATA_DTYPE(data)

        payload = FRAME_DTYPE(
            ((self.ax & Off_WF1F.AXON_MASK) << Off_WF1F.AXON_OFFSET)
            | ((self.ts & Off_WF1F.TIMESLOT_MASK) << Off_WF1F.TIMESLOT_OFFSET)
            | ((data & Off_WF1F.DATA_MASK) << Off_WF1F.DATA_OFFSET)
        )
        super().__init__(chip_coord, core_coord, rid, payload)

    @property
    def target_ts(self) -> int:
        return self.ts

    @property
    def target_axon(self) -> int:
        return self.ax

    @staticmethod
    @params_check(OffNeuDestInfoChecker)
    def _frame_dest_reorganized(dest_info: dict[str, Any]) -> FrameArrayType:
        return OfflineWorkFrame1.concat_frame_dest(
            (dest_info["addr_chip_x"], dest_info["addr_chip_y"]),
            (dest_info["addr_core_x"], dest_info["addr_core_y"]),
            (dest_info["addr_core_x_ex"], dest_info["addr_core_y_ex"]),
            dest_info["addr_axon"],
            dest_info["tick_relative"],
        )

    @classmethod
    def concat_frame_dest(
        cls,
        chip_coord: CoordLike,
        core_coord: CoordLike,
        rid: RIdLike,
        /,
        axons: ArrayLike,
        timeslots: ArrayLike | None = None,
    ) -> FrameArrayType:
        ax = np.asarray(axons, dtype=FRAME_DTYPE).ravel()

        if timeslots is None:
            ts = np.zeros_like(ax)
        else:
            ts = np.asarray(timeslots, dtype=FRAME_DTYPE).ravel()

        if ax.size != ts.size:
            raise ValueError(
                f"the size of axons & timeslots are not equal, {ax.size} != {ts.size}."
            )

        frame_dest = get_frame_dest(
            cls.header, to_coord(chip_coord), to_coord(core_coord), to_rid(rid)
        )

        payload = ((ax & Off_WF1F.AXON_MASK) << Off_WF1F.AXON_OFFSET) | (
            (ts & Off_WF1F.TIMESLOT_MASK) << Off_WF1F.TIMESLOT_OFFSET
        )

        return (frame_dest + payload).astype(FRAME_DTYPE)


class _WorkFrame2Base(Frame):
    header = FH.WORK_TYPE2

    def __init__(self, chip_coord: ChipCoord, n_sync: int) -> None:
        if n_sync > Off_WF2F.N_SYNC_MASK:
            warnings.warn(
                OUT_OF_RANGE_WARNING.format("n_sync", 30, n_sync), TruncationWarning
            )

        super().__init__(
            chip_coord,
            Coord(0, 0),
            _rid_unset(),
            FRAME_DTYPE(n_sync & Off_WF2F.N_SYNC_MASK),
        )


class _WorkFrame3Base(Frame):
    header = FH.WORK_TYPE3

    def __init__(self, chip_coord: ChipCoord) -> None:
        super().__init__(chip_coord, Coord(0, 0), _rid_unset())


class _WorkFrame4Base(Frame):
    header = FH.WORK_TYPE4

    def __init__(self, chip_coord: ChipCoord) -> None:
        super().__init__(chip_coord, Coord(0, 0), _rid_unset())


class OfflineWorkFrame2(_WorkFrame2Base):
    pass


class OfflineWorkFrame3(_WorkFrame3Base):
    pass


class OfflineWorkFrame4(_WorkFrame4Base):
    pass


class OnlineConfigFrame1(_LUTRAMFrame):
    header = FH.CONFIG_TYPE1


class OnlineConfigFrame2(_OnlineCoreRegFrame):
    header = FH.CONFIG_TYPE2


class OnlineConfigFrame3(_OnlineNeuRAMFrame):
    header = FH.CONFIG_TYPE3


class OnlineConfigFrame4(_OnlineWeightRAMFrame):
    header = FH.CONFIG_TYPE4


class OnlineTestInFrame1(Frame):
    header = FH.TEST_TYPE1

    def __init__(self, chip_coord: ChipCoord, core_coord: Coord, rid: RId) -> None:
        super().__init__(chip_coord, core_coord, rid)


class OnlineTestOutFrame1(_LUTRAMFrame):
    header = FH.TEST_TYPE1


class OnlineTestInFrame2(Frame):
    header = FH.TEST_TYPE2

    def __init__(self, chip_coord: ChipCoord, core_coord: Coord, rid: RId) -> None:
        super().__init__(chip_coord, core_coord, rid)


class OnlineTestOutFrame2(_OnlineCoreRegFrame):
    header = FH.TEST_TYPE2


class OnlineTestInFrame3(FramePackage):
    header = FH.TEST_TYPE3

    def __init__(
        self,
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        ram_start_addr: int,
        n_package: int,
    ) -> None:
        payload = FramePackagePayload(ram_start_addr, FPType.TESTIN, n_package)
        super().__init__(chip_coord, core_coord, rid, payload)


class OnlineTestOutFrame3(_OnlineNeuRAMFrame):
    header = FH.TEST_TYPE3


class OnlineTestInFrame4(FramePackage):
    header = FH.TEST_TYPE4

    def __init__(
        self,
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        ram_start_addr: int,
        n_package: int,
    ) -> None:
        payload = FramePackagePayload(ram_start_addr, FPType.TESTIN, n_package)
        super().__init__(chip_coord, core_coord, rid, payload)


class OnlineTestOutFrame4(_OnlineWeightRAMFrame):
    header = FH.TEST_TYPE4


class _OnlineWorkFrame1Base(Frame):
    header = FH.WORK_TYPE1

    def __init__(
        self,
        chip_coord: ChipCoord,
        coor_core: Coord,
        rid: RId,
        payload: FRAME_DTYPE = FRAME_DTYPE(0),
    ) -> None:
        if self.subtype is None:
            raise ValueError(f"'subtype' is not set for {self.__class__.__name__}.")

        _payload = FRAME_DTYPE(
            ((self.subtype & On_WF1F.SUBTYPE_MASK) << On_WF1F.SUBTYPE_OFFSET) + payload
        )
        super().__init__(chip_coord, coor_core, rid, _payload, subtype=self.subtype)


class OnlineWorkFrame1_1(_OnlineWorkFrame1Base):
    subtype = Online_WF1F_SubType.TYPE_I

    def __init__(
        self,
        chip_coord: ChipCoord,
        core_coord: Coord,
        rid: RId,
        timeslot: IntScalarType,
        axon: IntScalarType,
    ) -> None:
        self.ax = int(axon)
        self.ts = int(timeslot)

        if self.ts > On_WF1F_1.TIMESLOT_MASK or self.ts < 0:
            raise ValueError(
                f"timeslot out of range {On_WF1F_1.TIMESLOT_MASK} ({self.ts})."
            )

        if self.ax > OnCoreParams.ADDR_AXON_MAX or self.ax < 0:
            raise ValueError(
                f"axon out of range {OnCoreParams.ADDR_AXON_MAX} ({self.ax})."
            )

        payload = FRAME_DTYPE(
            ((axon & On_WF1F_1.AXON_MASK) << On_WF1F_1.AXON_OFFSET)
            | ((timeslot & On_WF1F_1.TIMESLOT_MASK) << On_WF1F_1.TIMESLOT_OFFSET)
        )
        super().__init__(chip_coord, core_coord, rid, payload)

    @property
    def target_ts(self) -> int:
        return self.ts

    @property
    def target_axon(self) -> int:
        return self.ax

    @staticmethod
    @params_check(OnNeuDestInfoChecker)
    def _frame_dest_reorganized(dest_info: dict[str, Any]) -> FrameArrayType:
        return OnlineWorkFrame1_1.concat_frame_dest(
            (dest_info["addr_chip_x"], dest_info["addr_chip_y"]),
            (dest_info["addr_core_x"], dest_info["addr_core_y"]),
            (dest_info["addr_core_x_ex"], dest_info["addr_core_y_ex"]),
            dest_info["addr_axon"],
            dest_info["tick_relative"],
        )

    @classmethod
    def concat_frame_dest(
        cls,
        chip_coord: CoordLike,
        core_coord: CoordLike,
        rid: RIdLike,
        /,
        axons: ArrayLike,
        timeslots: ArrayLike | None = None,
    ) -> FrameArrayType:
        ax = np.asarray(axons, dtype=FRAME_DTYPE).ravel()

        if timeslots is None:
            ts = np.zeros_like(ax)
        else:
            ts = np.asarray(timeslots, dtype=FRAME_DTYPE).ravel()

        if ax.size != ts.size:
            raise ValueError(
                f"the size of axons & timeslots are not equal, {ax.size} != {ts.size}."
            )

        frame_dest = get_frame_dest(
            cls.header, to_coord(chip_coord), to_coord(core_coord), to_rid(rid)
        )

        if cls.subtype is None:
            raise ValueError(f"'subtype' is not set for {cls.__name__}.")

        payload = (
            ((cls.subtype & On_WF1F_1.SUBTYPE_MASK) << On_WF1F_1.SUBTYPE_OFFSET)
            | ((ax & On_WF1F_1.AXON_MASK) << On_WF1F_1.AXON_OFFSET)
            | ((ts & On_WF1F_1.TIMESLOT_MASK) << On_WF1F_1.TIMESLOT_OFFSET)
        )

        return (frame_dest + payload).astype(FRAME_DTYPE)


class OnlineWorkFrame1_2(_OnlineWorkFrame1Base):
    subtype = Online_WF1F_SubType.TYPE_II


class OnlineWorkFrame1_3(_OnlineWorkFrame1Base):
    subtype = Online_WF1F_SubType.TYPE_III


class OnlineWorkFrame1_4(_OnlineWorkFrame1Base):
    subtype = Online_WF1F_SubType.TYPE_IV


class OnlineWorkFrame2(_WorkFrame2Base):
    pass


class OnlineWorkFrame3(_WorkFrame3Base):
    pass


class OnlineWorkFrame4(_WorkFrame4Base):
    pass



from importlib.metadata import version

# Coordinate definitions
from .coordinate import (
    Coord,
    OfflineCoord,
    OnlineCoord,
    CoordOffset,
    ReplicationId,
    CoordXY,
    CoordXYOffset,
    CoordZXYOffset,
    CoordXYUnitVec,
    ChipCoord,
    CoordAddr,
    CoordTuple2d,
    CoordTuple3d,
    CoordLike,
    RIdLike,
    CoordXYLike,
    CoordXYOffsetLike,
    CoordZXYOffsetLike,
    to_coord,
    to_coords,
    to_coordoffset,
    to_rid,
    to_coordxy,
    to_coordxys,
    to_coordxyoffset,
    to_coordzxyoffset,
)

# Chip v1
# Hardware constants
from .hw_defs import HwParams as HwConfig  # keep the compatibility
from .hw_defs import HwCoreParams as HwCoreCfg
from .hw_defs import HwOfflineCoreParams as OffCoreCfg
from .hw_defs import HwOnlineCoreParams as OnCoreCfg

HwCfg = HwConfig
# Chip v2.5
from .hw_defs import HwParamsV2 as HwConfigV2

# Frame library
from .framelib import (
    OfflineFrameGen,
    OnlineFrameGen,
    ChipFrameGen,
    FrameGenV2,
    OfflineFrameGenV2,
    OfflineConfigFrame1,
    OfflineConfigFrame2,
    OfflineConfigFrame3,
    OfflineConfigFrame4,
    OfflineWorkFrame1,
    OfflineWorkFrame2,
    OfflineWorkFrame3,
    OfflineWorkFrame4,
    OfflineTestInFrame1,
    OfflineTestInFrame2,
    OfflineTestInFrame3,
    OfflineTestInFrame4,
    OfflineTestOutFrame1,
    OfflineTestOutFrame2,
    OfflineTestOutFrame3,
    OfflineTestOutFrame4,
    OnlineConfigFrame1,
    OnlineConfigFrame2,
    OnlineConfigFrame3,
    OnlineConfigFrame4,
    OnlineWorkFrame1_1,
    OnlineWorkFrame1_2,
    OnlineWorkFrame1_3,
    OnlineWorkFrame1_4,
    OnlineWorkFrame2,
    OnlineWorkFrame3,
    OnlineWorkFrame4,
    OnlineTestInFrame1,
    OnlineTestInFrame2,
    OnlineTestInFrame3,
    OnlineTestInFrame4,
    OnlineTestOutFrame1,
    OnlineTestOutFrame2,
    OnlineTestOutFrame3,
    OnlineTestOutFrame4,
    FRAME_DTYPE,
    LUT_DTYPE,
    PAYLOAD_DATA_DTYPE,
    FrameArrayType,
    LUTDataType,
    PayloadDataType,
    LUTPotentialType,
    LUT_POTENTIAL_DTYPE,
    LUTActivationType,
    LUT_ACTIVATION_DTYPE,
)

# Routing
from .routing_defs import (
    RoutingCoord,
    RoutingDirection,
    RoutingLevel,
    RoutingPath,
    RoutingStatus,
    get_replication_id,
    get_multicast_cores,
    ONLINE_CORES_BASE_COORD,
    ROUTING_DIRECTIONS_IDX,
)

# Core register definitions & parameters model
from .core_model import CoreReg, OfflineCoreReg, OnlineCoreReg
from .core_defs import (
    CoreRegLim,
    OfflineCoreRegLim,
    OnlineCoreRegLim,
    WeightWidth,
    LCN_EX,
    InputWidthFormat,
    SpikeWidthFormat,
    MaxPoolingEnable,
    SNNModeEnable,
    LUTRandomEnable,
    DecayRandomEnable,
    LeakOrder,
    OnlineModeEnable,
    CoreType,
    CoreMode,
    get_core_mode,
    core_mode_check,
)

# Neuron register definitions & parameters model
from .neuron_model import (
    NeuDestInfo,
    OfflineNeuDestInfo,
    OnlineNeuDestInfo,
    NeuAttrs,
    OfflineNeuAttrs,
    OnlineNeuAttrs,
    OfflineNeuConf,
    OnlineNeuConf,
)
from .neuron_defs import (
    NeuRegLim,
    OfflineNeuRegLim,
    OnlineNeuRegLim,
    OnlineNeuRegLim_WW1,
    OnlineNeuRegLim_WWn,
    SynapticIntegrationMode as SIM,
    ResetMode as RM,
    NegativeThresholdMode as NTM,
    LeakIntegrationMode as LIM,
    LeakDirectionMode as LDM,
    LeakComparisonMode as LCM,
)

# Chip v2.5
# Core register definitions & parameters model
from .core_model_v2 import OfflineCoreRegV2
from .core_defs_v2 import (
    AddPotentialMode,
    CSCAccelerateMode,
    DataSign,
    DataWidth,
    OfflineCoreRegLimV2,
    PoolingMode,
    SNNMode,
    ZeroOutputMode,
)

# Neuron register definitions & parameters model
from .neuron_model_v2 import (
    NeuDestInfoV2,
    OfflineNeuDestInfoV2,
    OfflineNeuFullAttrsV2Part1,
    OfflineNeuFullAttrsV2Part2,
    OfflineNeuHalfAttrsV2,
    OfflineNeuFullAttrsV2,
    OfflineNeuFoldedAttrsV2Part1,
    OfflineNeuFoldedAttrsV2Part2,
    OnlineNeuFoldedAttrsV2Part1,
    OnlineNeuFoldedAttrsV2Part2,
    OfflineNeuFullConfV2,
    OfflineNeuHalfConfV2,
)
from .neuron_defs_v2 import (
    OutputType,
    FoldType,
    NeuronType,
    ThresholdNegMode,
    ThresholdPosMode,
    LateralInhibitionMode,
    LeakMultiComparisonOrder,
    LeakMultiInputMode,
    LeakMultiMode,
    LeakAddMode,
    WeightCompressType,
    OfflineNeuRegLimV2,
)

# Routing
from .routing_hexa import (
    AERPacketZXYCopy,
    AERPacket,
    aer_packet_walk,
    aer_packet_area,
    find_coordxy_shortest_path,
)

try:
    __version__ = version("paicorelib")
except Exception:
    __version__ = None



from enum import IntEnum, unique

from .utils import _mask

__all__ = [
    # Core reg limits
    "OfflineCoreRegLimV2",
    # Types
    "SNNMode",
    "PoolingMode",
    "AddPotentialMode",
    "ZeroOutputMode",
    "DataSign",
    "DataWidth",
    "CSCAccelerateMode",
]


class OfflineCoreRegLimV2:
    """Limits of offline core registers for chip v2.5."""

    AXON_SKEW_MIN = -_mask(15)
    AXON_SKEW_MAX = -AXON_SKEW_MIN
    NEURON_NUMBER_MAX = _mask(12) + 1  # <= 4096
    TEST_CORE_COORD_MIN = -_mask(5)
    TEST_CORE_COORD_MAX = -TEST_CORE_COORD_MIN
    GLOBAL_SEND_MAX = _mask(7)
    GLOBAL_RECEIVE_MAX = _mask(6)
    THREAD_NUMBER_MAX = _mask(10)
    BUSY_CYCLE_MAX = _mask(12)
    DELAY_CYCLE_MAX = _mask(16)
    WIDTH_CYCLE_MAX = _mask(8)
    TICK_START_MAX = _mask(16)
    TICK_DURATION_MAX = _mask(32)
    TICK_INITIAL_MAX = _mask(16)


@unique
class SNNMode(IntEnum):
    """SNN and ANN mode selection.
    0: SNN mode, using LIF operation rules
    1: ANN mode, using activation function operation rules
    """

    SNN = 0
    ANN = 1


@unique
class PoolingMode(IntEnum):
    """Pooling mode selection.
    0: Average pooling
    1: Max pooling
    """

    AVERAGE = 0
    MAX = 1


@unique
class AddPotentialMode(IntEnum):
    """Accumulation mode selection.
    0: Normal accumulation
    1: Direct accumulation of membrane potential
    """

    NORMAL = 0
    DIRECT_ADD = 1


@unique
class ZeroOutputMode(IntEnum):
    """Whether to output zero values.
    0: Do not output zero values
    1: Output zero values
    """

    DISABLE = 0
    ENABLE = 1


@unique
class DataWidth(IntEnum):
    """Input/output/weight data width."""

    WIDTH_1BIT = 0
    WIDTH_2BIT = 1
    WIDTH_4BIT = 2
    WIDTH_8BIT = 3
    WIDTH_16BIT = 4
    WIDTH_32BIT = 5

@unique
class DataSign(IntEnum):
    """Input/output/weight data sign."""

    UNSIGNED = 0
    SIGNED = 1


@unique
class CSCAccelerateMode(IntEnum):
    """CSC compressed calculation acceleration mode.
    0: No acceleration, compressed storage calculation is slow
    1: Acceleration, compressed storage calculation is faster, but vjt_initial needs to store weight_address_start again
    """

    DISABLE = 0
    ENABLE = 1



from typing import Annotated

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    NonNegativeInt,
    TypeAdapter,
    model_validator,
)

from .neuron_defs import ResetMode
from .neuron_defs_v2 import (
    FoldType,
    LateralInhibitionMode,
    LeakAddMode,
    LeakMultiComparisonOrder,
    LeakMultiInputMode,
    LeakMultiMode,
    NeuronType,
    OfflineNeuRegLimV2,
    OutputType,
    ThresholdNegMode,
    ThresholdPosMode,
    WeightCompressType,
)
from .neuron_model import NeuAttrs  # Reuse the base model

__all__ = [
    "NeuDestInfoV2",
    "OfflineNeuDestInfoV2",
    "OfflineNeuFullAttrsV2",
    "OfflineNeuFullAttrsV2Part1",
    "OfflineNeuFullAttrsV2Part2",
    "OfflineNeuHalfAttrsV2",
    "OfflineNeuFoldedAttrsV2Part1",
    "OfflineNeuFoldedAttrsV2Part2",
    "OnlineNeuFoldedAttrsV2Part1",
    "OnlineNeuFoldedAttrsV2Part2",
    "OfflineNeuFullConfV2",
    "OfflineNeuHalfConfV2",
]


class NeuDestInfoV2(BaseModel):
    model_config = ConfigDict(
        extra="ignore", validate_assignment=True, use_enum_values=True
    )

    tick_relative: Annotated[
        NonNegativeInt,
        Field(
            le=OfflineNeuRegLimV2.TICK_RELATIVE_MAX,
            description="Relative time information.",
        ),
    ]
    addr_axon: Annotated[
        NonNegativeInt,
        Field(le=OfflineNeuRegLimV2.ADDR_AXON_MAX, description="Target axon address."),
    ]
    addr_core_xy: Annotated[
        int,
        Field(
            ge=OfflineNeuRegLimV2.ADDR_CORE_COORD_MIN,
            le=OfflineNeuRegLimV2.ADDR_CORE_COORD_MAX,
            description="Target core relative XY address.",
        ),
    ]
    addr_core_x: Annotated[
        int,
        Field(
            ge=OfflineNeuRegLimV2.ADDR_CORE_COORD_MIN,
            le=OfflineNeuRegLimV2.ADDR_CORE_COORD_MAX,
            description="Target core relative X address.",
        ),
    ]
    addr_core_y: Annotated[
        int,
        Field(
            ge=OfflineNeuRegLimV2.ADDR_CORE_COORD_MIN,
            le=OfflineNeuRegLimV2.ADDR_CORE_COORD_MAX,
            description="Target core relative Y address.",
        ),
    ]
    addr_copy_xy: Annotated[
        int,
        Field(
            ge=OfflineNeuRegLimV2.ADDR_CORE_COORD_MIN,
            le=OfflineNeuRegLimV2.ADDR_CORE_COORD_MAX,
            description="Number of copies in XY direction.",
        ),
    ]
    addr_copy_x: Annotated[
        int,
        Field(
            ge=OfflineNeuRegLimV2.ADDR_CORE_COORD_MIN,
            le=OfflineNeuRegLimV2.ADDR_CORE_COORD_MAX,
            description="Number of copies in X direction.",
        ),
    ]
    addr_copy_y: Annotated[
        int,
        Field(
            ge=OfflineNeuRegLimV2.ADDR_CORE_COORD_MIN,
            le=OfflineNeuRegLimV2.ADDR_CORE_COORD_MAX,
            description="Number of copies in Y direction.",
        ),
    ]


class NeuCommonAttrsV2(NeuAttrs):
    weight_skew: Annotated[
        NonNegativeInt,
        Field(
            le=OfflineNeuRegLimV2.WEIGHT_SKEW_MAX,
            description="Vertical offset of neuron corresponding weight.",
        ),
    ]
    weight_address_start: Annotated[
        NonNegativeInt,
        Field(
            le=OfflineNeuRegLimV2.WEIGHT_ADDRESS_MAX,
            description="Start address of neuron corresponding weight SRAM.",
        ),
    ]
    weight_address_end: Annotated[
        NonNegativeInt,
        Field(
            le=OfflineNeuRegLimV2.WEIGHT_ADDRESS_MAX,
            description="End address of neuron corresponding weight SRAM.",
        ),
    ]

    fold_type: Annotated[FoldType, Field(description="Fold type selection.")]
    neuron_type: Annotated[NeuronType, Field(description="Neuron type selection.")]


class OfflineNeuDestInfoV2(NeuDestInfoV2):
    pass


class OfflineNeuCommonAttrsV2(NeuCommonAttrsV2):
    output_type: Annotated[OutputType, Field(description="Output type selection.")]
    vjt: Annotated[
        int, Field(default=0, description="Current time step membrane potential.")
    ] = 0


OfflineNeuHalfAttrsV2 = OfflineNeuCommonAttrsV2
OfflineNeuFullAttrsV2Part1 = OfflineNeuHalfAttrsV2


class OfflineNeuFullAttrsV2Part2(NeuAttrs):
    reset_mode: Annotated[ResetMode, Field(description="Reset mode selection.")]
    reset_v: Annotated[
        int,
        Field(
            ge=OfflineNeuRegLimV2.RESET_V_MIN,
            le=OfflineNeuRegLimV2.RESET_V_MAX,
            description="Membrane potential reset value.",
        ),
    ]

    threshold_neg_mode: Annotated[
        ThresholdNegMode, Field(description="Negative threshold mode selection.")
    ]
    threshold_pos_mode: Annotated[
        ThresholdPosMode, Field(description="Positive threshold mode selection.")
    ]

    threshold_neg: Annotated[int, Field(description="Negative threshold.")]
    threshold_pos: Annotated[int, Field(description="Positive threshold.")]

    lateral_inhibition: Annotated[
        LateralInhibitionMode, Field(description="Lateral inhibition mode selection.")
    ]
    leak_multi_sequence: Annotated[
        LeakMultiComparisonOrder, Field(description="Multiplicative leak sequence.")
    ]
    leak_multi_input: Annotated[
        LeakMultiInputMode,
        Field(description="Whether input participates in multiplicative leak."),
    ]
    leak_multi_mode: Annotated[
        LeakMultiMode, Field(description="Multiplicative leak mode selection.")
    ]
    leak_add_mode: Annotated[
        LeakAddMode, Field(description="Additive leak mode selection.")
    ]

    leak_tau: Annotated[
        int,
        Field(
            ge=OfflineNeuRegLimV2.LEAK_TAU_MIN,
            le=OfflineNeuRegLimV2.LEAK_TAU_MAX,
            description="Multiplicative leak shift amount.",
        ),
    ]
    leak_v: Annotated[
        int,
        Field(
            ge=OfflineNeuRegLimV2.LEAK_V_MIN,
            le=OfflineNeuRegLimV2.LEAK_V_MAX,
            description="Additive leak potential.",
        ),
    ]

    weight_compress: Annotated[
        WeightCompressType,
        Field(
            default=WeightCompressType.DENSE,
            description="Weight compression type (Dense/Sparse).",
        ),
    ]

    vjt_initial: Annotated[
        int,
        Field(
            default=0,
            ge=OfflineNeuRegLimV2.VJT_INITIAL_MIN,
            le=OfflineNeuRegLimV2.VJT_INITIAL_MAX,
            description="Initial membrane potential.",
        ),
    ] = 0


class OfflineNeuFullAttrsV2(OfflineNeuFullAttrsV2Part1, OfflineNeuFullAttrsV2Part2):
    pass


class NeuFoldedAttrsV2Part1(NeuAttrs):
    fold_range_xy: Annotated[
        NonNegativeInt,
        Field(
            le=OfflineNeuRegLimV2.FOLD_RANGE_MAX, description="XY dimension fold width."
        ),
    ]
    fold_range_x: Annotated[
        NonNegativeInt,
        Field(
            le=OfflineNeuRegLimV2.FOLD_RANGE_MAX, description="X dimension fold width."
        ),
    ]
    fold_range_y: Annotated[
        NonNegativeInt,
        Field(
            le=OfflineNeuRegLimV2.FOLD_RANGE_MAX, description="Y dimension fold width."
        ),
    ]

    fold_skew_xy: Annotated[
        NonNegativeInt,
        Field(
            le=OfflineNeuRegLimV2.FOLD_SKEW_MAX,
            description="XY dimension weight offset step.",
        ),
    ]
    fold_skew_x: Annotated[
        NonNegativeInt,
        Field(
            le=OfflineNeuRegLimV2.FOLD_SKEW_MAX,
            description="X dimension weight offset step.",
        ),
    ]
    fold_skew_y: Annotated[
        NonNegativeInt,
        Field(
            le=OfflineNeuRegLimV2.FOLD_SKEW_MAX,
            description="Y dimension weight offset step.",
        ),
    ]

    fold_axon_xy: Annotated[
        NonNegativeInt,
        Field(
            le=OfflineNeuRegLimV2.FOLD_AXON_MAX,
            description="XY dimension Axon address offset step.",
        ),
    ]
    fold_axon_x: Annotated[
        NonNegativeInt,
        Field(
            le=OfflineNeuRegLimV2.FOLD_AXON_MAX,
            description="X dimension Axon address offset step.",
        ),
    ]
    fold_axon_y: Annotated[
        NonNegativeInt,
        Field(
            le=OfflineNeuRegLimV2.FOLD_AXON_MAX,
            description="Y dimension Axon address offset step.",
        ),
    ]

    fold_number: Annotated[
        NonNegativeInt,
        Field(
            le=OfflineNeuRegLimV2.FOLD_NUMBER_MAX,
            description="Total number of folded neurons.",
        ),
    ]

    @model_validator(mode="after")
    def check_fold_number(self):
        if (
            self.fold_range_xy * self.fold_range_x * self.fold_range_y
            != self.fold_number
        ):
            raise ValueError(
                "'fold_number' must be equal to 'fold_range_xy' * 'fold_range_x' * 'fold_range_y', "
                f"but {self.fold_number} != {self.fold_range_xy} * {self.fold_range_x} * {self.fold_range_y}"
            )

        return self


class OfflineNeuFoldedAttrsV2Part1(NeuFoldedAttrsV2Part1):
    pass


class OnlineNeuFoldedAttrsV2Part1(NeuFoldedAttrsV2Part1):
    pass


class OfflineNeuFoldedAttrsV2Part2(NeuAttrs):
    fold_vjt_3: Annotated[int, Field(description="Folded neuron 3 membrane potential.")]
    fold_vjt_2: Annotated[int, Field(description="Folded neuron 2 membrane potential.")]
    fold_vjt_1: Annotated[int, Field(description="Folded neuron 1 membrane potential.")]
    fold_vjt_0: Annotated[int, Field(description="Folded neuron 0 membrane potential.")]


class OnlineNeuFoldedAttrsV2Part2(NeuAttrs):
    fold_vjt_3: Annotated[
        float, Field(description="Folded neuron 3 membrane potential.")
    ]
    fold_vjt_2: Annotated[
        float, Field(description="Folded neuron 2 membrane potential.")
    ]
    fold_vjt_1: Annotated[
        float, Field(description="Folded neuron 1 membrane potential.")
    ]
    fold_vjt_0: Annotated[
        float, Field(description="Folded neuron 0 membrane potential.")
    ]


# Neuron configuration
class OfflineNeuFullConfV2(BaseModel):
    attrs: OfflineNeuFullAttrsV2
    dest_info: OfflineNeuDestInfoV2


class OfflineNeuHalfConfV2(BaseModel):
    attrs: OfflineNeuHalfAttrsV2
    dest_info: OfflineNeuDestInfoV2


OfflineNeuDestInfoV2Checker = TypeAdapter(OfflineNeuDestInfoV2)
OfflineNeuHalfAttrsV2Checker = TypeAdapter(OfflineNeuHalfAttrsV2)
OfflineNeuFullAttrsV2Part2Checker = TypeAdapter(OfflineNeuFullAttrsV2Part2)
OfflineNeuFoldedAttrsV2Part1Checker = TypeAdapter(OfflineNeuFoldedAttrsV2Part1)
OfflineNeuFoldedAttrsV2Part2Checker = TypeAdapter(OfflineNeuFoldedAttrsV2Part2)



import logging
import sys

log_level_to_abbr = {
    "DEBUG": "D",
    "INFO": "I",
    "WARNING": "W",
    "ERROR": "E",
    "CRITICAL": "C",
}


class PAIlibLogsFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord) -> str:
        s = super().format(record)
        record.asctime = self.formatTime(record, "%H:%M:%S")

        shortlevel = log_level_to_abbr.get(record.levelname, record.levelname)
        prefix = f"[{shortlevel} {record.asctime} {record.lineno}]{record.module}"
        lines = s.split("\n")
        return "\n".join(f"{prefix} {li}" for li in lines)


DEFAULT_FORMATTER = PAIlibLogsFormatter()
_logging_enabled = False


def setup_logger(name: str) -> logging.Logger:
    logger = logging.getLogger(name)

    if not logger.handlers:
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(DEFAULT_FORMATTER)
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)

    logger.disabled = not _logging_enabled
    return logger


def enable_logging(level: int = logging.DEBUG) -> None:
    global _logging_enabled
    _logging_enabled = True

    for name in logging.Logger.manager.loggerDict:
        logger = logging.getLogger(name)
        logger.disabled = False
        logger.setLevel(level)


def disable_logging() -> None:
    global _logging_enabled
    _logging_enabled = False

    for name in logging.Logger.manager.loggerDict:
        logger = logging.getLogger(name)
        logger.disabled = True


def get_logger(name: str) -> logging.Logger:
    logger = setup_logger(name)
    logger.disabled = not _logging_enabled
    return logger



from abc import ABCMeta
from dataclasses import dataclass, field
from typing import ClassVar

import numpy as np

from ..coordinate import ChipCoord, Coord, CoordZXYOffset
from ..coordinate import ReplicationId as RId
from ..routing_hexa import AERPacketZXYCopy
from .frame_defs import FF, FFV2, FT, FramePackageType, Online_WF1F_SubType
from .frame_defs import FrameHeader as FH
from .frame_defs import OfflineNeuRAMFormat as Off_NRAMF
from .types import FRAME_DTYPE, FrameArrayType
from .utils import header2type

__all__ = [
    "Frame",
    "FramePackage",
    "FramePackagePayload",
    "FramePackagePayloadV2",
    "FrameV2",
    "FramePackageHeaderV2",
    "get_frame_dest",
    "get_frame_destV2",
]


@dataclass
class FramePackagePayload:
    """The payload of a frame package."""

    ram_start_addr: int
    """Neuron start address."""
    package_type: FramePackageType
    """Type of the package, config/test-in or test-out."""
    n_package: int = 0
    """#N of packages on the SRAM."""

    def __post_init__(self) -> None:
        if (
            self.ram_start_addr > Off_NRAMF.GENERAL_PACKAGE_NEU_START_ADDR_MASK
            or self.ram_start_addr < 0
        ):
            raise ValueError(
                f"neuron base address out of range [0, {Off_NRAMF.GENERAL_PACKAGE_NEU_START_ADDR_MASK}], "
                f"got {self.ram_start_addr}."
            )

        if self.n_package > Off_NRAMF.GENERAL_PACKAGE_NUM_MASK or self.n_package < 0:
            raise ValueError(
                f"the number of data packages out of range [0, {Off_NRAMF.GENERAL_PACKAGE_NUM_MASK}], "
                f"got {self.n_package}."
            )

    @property
    def value(self) -> FRAME_DTYPE:
        return FRAME_DTYPE(
            (
                (self.ram_start_addr & FF.GENERAL_PACKAGE_NEU_START_ADDR_MASK)
                << FF.GENERAL_PACKAGE_NEU_START_ADDR_OFFSET
            )
            | (
                (self.package_type & FF.GENERAL_PACKAGE_TYPE_MASK)
                << FF.GENERAL_PACKAGE_TYPE_OFFSET
            )
            | (
                (self.n_package & FF.GENERAL_PACKAGE_NUM_MASK)
                << FF.GENERAL_PACKAGE_NUM_OFFSET
            )
        )

    def __str__(self) -> str:
        return f"start_addr={self.ram_start_addr}, type={self.package_type.name}, n={self.n_package}"


def get_frame_dest(
    header: FH, chip_coord: ChipCoord, core_coord: Coord, rid: RId
) -> int:
    h = header.value & FF.GENERAL_HEADER_MASK
    chip_addr = chip_coord.address & FF.GENERAL_CHIP_ADDR_MASK
    core_addr = core_coord.address & FF.GENERAL_CORE_ADDR_MASK
    rid_addr = rid.address & FF.GENERAL_CORE_EX_ADDR_MASK

    return (
        (h << FF.GENERAL_HEADER_OFFSET)
        + (chip_addr << FF.GENERAL_CHIP_ADDR_OFFSET)
        + (core_addr << FF.GENERAL_CORE_ADDR_OFFSET)
        + (rid_addr << FF.GENERAL_CORE_EX_ADDR_OFFSET)
    )


class FrameMeta(ABCMeta):
    def __call__(cls, *args, **kwargs):
        if cls.__name__ in ("Frame", "FramePackage", "_FrameBase"):
            header_override = kwargs.pop("header", None)
            instance = super().__call__(*args, **kwargs)

            if header_override is not None:
                instance.header = header_override
            return instance
        else:
            return super().__call__(*args, **kwargs)


@dataclass
class _FrameBase(metaclass=FrameMeta):
    """Frame common part:
    [Header] + [chip coordinate] + [core coordinate] + [replication id] + [payload]
     4 bits         10 bits             10 bits             10 bits        30 bits
    """

    header: ClassVar[FH]
    chip_coord: ChipCoord
    core_coord: Coord
    rid: RId
    subtype: Online_WF1F_SubType | None = field(default=None, kw_only=True)

    @property
    def frame_type(self) -> FT:
        return header2type(self.header)

    @property
    def frame_dest(self) -> int:
        return get_frame_dest(self.header, self.chip_coord, self.core_coord, self.rid)


@dataclass
class Frame(_FrameBase):
    """Frames which contains information.

    1. Single frame:
        [Common part] + [payload]
           30 bits       30 bits

    2. Frames group = N * single frame:
        [Common part] + [payload[0]]
        [Common part] + [payload[1]]
        [Common part] + [payload[2]]
                    ...
           30 bits        30 bits
    """

    _payload: np.unsignedinteger | FrameArrayType = field(default=FRAME_DTYPE(0))

    @property
    def value(self) -> FrameArrayType:
        """Get the full frames of the frame."""
        value = self.frame_dest + (self.payload & FF.GENERAL_PAYLOAD_MASK)
        value = np.asarray(value, dtype=FRAME_DTYPE)
        value.setflags(write=False)
        return value

    @property
    def payload(self) -> FrameArrayType:
        """Get the payload of the frame."""
        return np.atleast_1d(self._payload).view(FRAME_DTYPE)

    def __len__(self) -> int:
        return 1 if isinstance(self.payload, int) else self.payload.size

    def __str__(self) -> str:
        return (
            f"Frame text:\n"
            f"Head:             {self.header}\n"
            f"Chip coord:       {self.chip_coord}\n"
            f"Core coord:       {self.core_coord}\n"
            f"Replication id:   {self.rid}\n"
            f"Payload:          {self.payload}\n"
        )

    def _make_same_dest(self, payload: FRAME_DTYPE | FrameArrayType):
        """Make a new frame with the same destination as the current frame."""
        return type(self)(self.chip_coord, self.core_coord, self.rid, payload)


@dataclass
class FramePackage(_FrameBase):
    """Frame package for a length of `N` contents:

    1. [Header(sub type)] + [chip coordinate] + [core coordinate] + [replication id] + [payload]
            4 bits               10 bits            10 bits            10 bits          30 bits
    2. [contents[0]], 64 bits.
    ...
    N+1. [contents[N-1]], 64 bits.

    NOTE: N can be 0. At this case, the frame package is empty.
    """

    PACKAGE_DATA_SHOW_MAX_LINE: ClassVar[int] = 99

    _payload: FramePackagePayload
    packages: FrameArrayType = field(
        default_factory=lambda: np.zeros(0, dtype=FRAME_DTYPE)
    )

    @property
    def payload(self) -> FramePackagePayload:
        """Get the payload of the frame package."""
        return self._payload

    @property
    def n_package(self) -> int:
        return self.packages.size

    @property
    def value(self) -> FrameArrayType:
        """Get the full frames of the group."""
        value = np.zeros((len(self),), dtype=FRAME_DTYPE)

        value[0] = self.frame_dest + (
            self.payload.value & FRAME_DTYPE(FF.GENERAL_PAYLOAD_MASK)
        )
        if self.n_package > 0:
            value[1:] = self.packages.copy()

        value.setflags(write=False)
        return value

    def __len__(self) -> int:
        return 1 + self.n_package

    def __str__(self) -> str:
        text = "Frame package info:"
        text += f"\nheader:         {self.header}"
        text += f"\nchip coord:      {self.chip_coord}"
        text += f"\ncore coord:      {self.core_coord}"
        text += f"\nreplication id:  {self.rid}"
        text += f"\npayload:         {self.payload}"

        if self.n_package > 0:
            text += "\nPackage data:"

            for i in range(self.n_package):
                text += f"#{i:2d}: {self.packages[i]}\n"
                if i >= self.PACKAGE_DATA_SHOW_MAX_LINE:
                    text += (
                        f"... (showing {self.PACKAGE_DATA_SHOW_MAX_LINE} lines only)\n"
                    )
                    break

        return text

    def _make_same_dest(
        self, payload: FramePackagePayload, packages: FrameArrayType
    ) -> "FramePackage":
        """Make a new frame with the same destination as the current frame package."""
        return type(self)(self.chip_coord, self.core_coord, self.rid, payload, packages)


def get_frame_destV2(
    header: FH, pkt_offset: CoordZXYOffset, pkt_ncopy: AERPacketZXYCopy
) -> int:
    oz, ox, oy = pkt_offset.to_sign_magnitude()
    nz, nx, ny = pkt_ncopy.to_sign_magnitude()
    pkt_addr = (
        (
            ((oz & FFV2.GENERAL_CORE_XY_ADDR_MASK) << FFV2.GENERAL_CORE_XY_ADDR_OFFSET)
            | ((ox & FFV2.GENERAL_CORE_X_ADDR_MASK) << FFV2.GENERAL_CORE_X_ADDR_OFFSET)
            | ((oy & FFV2.GENERAL_CORE_Y_ADDR_MASK) << FFV2.GENERAL_CORE_Y_ADDR_OFFSET)
        )
        | ((nz & FFV2.GENERAL_COPY_XY_ADDR_MASK) << FFV2.GENERAL_COPY_XY_ADDR_OFFSET)
        | ((nx & FFV2.GENERAL_COPY_X_ADDR_MASK) << FFV2.GENERAL_COPY_X_ADDR_OFFSET)
        | ((ny & FFV2.GENERAL_COPY_Y_ADDR_MASK) << FFV2.GENERAL_COPY_Y_ADDR_OFFSET)
    )
    return (
        (header.value & FFV2.GENERAL_HEADER_MASK) << FFV2.GENERAL_HEADER_OFFSET
    ) | pkt_addr


@dataclass
class _FrameBaseV2:
    header: FH
    pkt_offset: CoordZXYOffset
    pkt_ncopy: AERPacketZXYCopy

    def _pkt_addr_in_sign_magnitude(self) -> int:
        oz, ox, oy = self.pkt_offset.to_sign_magnitude()
        nz, nx, ny = self.pkt_ncopy.to_sign_magnitude()
        return (
            (
                (
                    (oz & FFV2.GENERAL_CORE_XY_ADDR_MASK)
                    << FFV2.GENERAL_CORE_XY_ADDR_OFFSET
                )
                | (
                    (ox & FFV2.GENERAL_CORE_X_ADDR_MASK)
                    << FFV2.GENERAL_CORE_X_ADDR_OFFSET
                )
                | (
                    (oy & FFV2.GENERAL_CORE_Y_ADDR_MASK)
                    << FFV2.GENERAL_CORE_Y_ADDR_OFFSET
                )
            )
            | (
                (nz & FFV2.GENERAL_COPY_XY_ADDR_MASK)
                << FFV2.GENERAL_COPY_XY_ADDR_OFFSET
            )
            | ((nx & FFV2.GENERAL_COPY_X_ADDR_MASK) << FFV2.GENERAL_COPY_X_ADDR_OFFSET)
            | ((ny & FFV2.GENERAL_COPY_Y_ADDR_MASK) << FFV2.GENERAL_COPY_Y_ADDR_OFFSET)
        )

    @property
    def frame_dest(self) -> int:
        return get_frame_destV2(self.header, self.pkt_offset, self.pkt_ncopy)

    def __str__(self) -> str:
        text = "Frame info:"
        text += f"\nheader:     {self.header}"
        text += f"\npkg offset: {self.pkt_offset}"
        text += f"\npkg ncopy:  {self.pkt_ncopy}"
        return text


class FramePackagePayloadV2(FramePackagePayload):
    def __post_init__(self) -> None:
        if (
            self.ram_start_addr > FFV2.GENERAL_PACKAGE_NEU_START_ADDR_MASK
            or self.ram_start_addr < 0
        ):
            raise ValueError(
                f"neuron base address out of range [0, {FFV2.GENERAL_PACKAGE_NEU_START_ADDR_MASK}], "
                f"got {self.ram_start_addr}."
            )

        if self.n_package > FFV2.GENERAL_PACKAGE_NUM_MASK or self.n_package < 0:
            raise ValueError(
                f"the number of data packages out of range [0, {FFV2.GENERAL_PACKAGE_NUM_MASK}], "
                f"got {self.n_package}."
            )

    @property
    def value(self) -> FRAME_DTYPE:
        return FRAME_DTYPE(
            (
                (self.ram_start_addr & FFV2.GENERAL_PACKAGE_NEU_START_ADDR_MASK)
                << FFV2.GENERAL_PACKAGE_NEU_START_ADDR_OFFSET
            )
            | (
                (self.package_type & FFV2.GENERAL_PACKAGE_TYPE_MASK)
                << FFV2.GENERAL_PACKAGE_TYPE_OFFSET
            )
            | (
                (self.n_package & FFV2.GENERAL_PACKAGE_NUM_MASK)
                << FFV2.GENERAL_PACKAGE_NUM_OFFSET
            )
        )


@dataclass
class FrameV2(_FrameBaseV2):
    payload: int

    @property
    def value(self) -> FrameArrayType:
        """Get the full frames of the frame."""
        frame = self.frame_dest | (
            (int(self.payload) & FFV2.GENERAL_PAYLOAD_MASK)
            << FFV2.GENERAL_PAYLOAD_OFFSET
        )
        return np.array([frame], dtype=FRAME_DTYPE)


@dataclass
class FramePackageHeaderV2(_FrameBaseV2):
    payload: FramePackagePayloadV2

    @classmethod
    def make_pkg_header(
        cls,
        header: FH,
        pkt_offset: CoordZXYOffset,
        pkt_ncopy: AERPacketZXYCopy,
        start_addr: int,
        pkg_type: FramePackageType,
        n_package: int = 0,
    ):
        return cls(
            header,
            pkt_offset,
            pkt_ncopy,
            FramePackagePayloadV2(start_addr, pkg_type, n_package),
        )

    @property
    def value(self) -> FrameArrayType:
        """Get the full frames of the frame package header."""
        frame = self.frame_dest | (
            (int(self.payload.value) & FFV2.GENERAL_PAYLOAD_MASK)
            << FFV2.GENERAL_PAYLOAD_OFFSET
        )
        return np.array([frame], dtype=FRAME_DTYPE)

    def __str__(self) -> str:
        return super().__str__() + f"\npayload:    {self.payload}"
